from dataclasses import dataclass
from typing import Callable, Optional, Sequence

import pyarrow as pa


@dataclass
class ModelSpecs:
    default_dimensions: int
    validate_dimensions: Callable[[int], bool]
    validation_message: str
    requires_none_input: bool = False


def create_fixedsize_with_nulls(
    vectors: Sequence[Optional[Sequence[float]]], vector_size: int
) -> pa.FixedSizeListArray:
    flat_values = []
    for vec in vectors:
        if vec is not None:
            flat_values.extend(vec)
        else:
            flat_values.extend([0] * vector_size)  # placeholder values

    # Create mask as PyArrow boolean array (True means null)
    mask = pa.array([vec is None for vec in vectors], type=pa.bool_())

    # Create the FixedSizeListArray with the mask
    return pa.FixedSizeListArray.from_arrays(pa.array(flat_values), vector_size, mask=mask)



from __future__ import annotations

import contextlib
from contextvars import ContextVar, Token
from typing import Dict, Optional, Union

JsonValue = Union[str, int, float, bool, None]
ContextJsonDict = Dict[str, JsonValue]


class ChalkContext:
    """
    An immutable context that can be accessed from Python resolvers.
    This context wraps a JSON-compatible dictionary or JSON string with type restrictions.

    Examples
    --------
    >>> from chalk.client import ChalkClient
    >>> from chalk.features import features
    >>> from chalk import ChalkContext, online
    >>> import requests
    >>> import json
    ...
    >>> @features
    ... class User:
    ...     id: int
    ...     endpoint_url: str
    ...     endpoint_response: str
    >>> @online
    ... def get_user_endpoint_response(endpoint_url: User.endpoint_url) -> User.endpoint_response:
    ...     context_headers = {}
    ...     optional_correlation_id = ChalkContext.get("request_correlation_id")
    ...     if optional_correlation_id is not None:
    ...         context_headers["correlation-id"] = optional_correlation_id
    ...     response = requests.get(endpoint_url, headers=context_headers)
    ...     return json.dumps(response.json())
    >>> ChalkClient().query(
    ...     input={User.id: 1, User.endpoint_url: "https://api.example.com/message"},
    ...     output=[User.endpoint_response],
    ...     query_context={"request_correlation_id": "df0cc84b-bb0e-41b1-82cd-74ccd968b2fa"},
    ... )
    """

    _context_var: ContextVar[ContextJsonDict | None] = ContextVar("context_dict", default=None)

    @classmethod
    @contextlib.contextmanager
    def _set_context(cls, context: ContextJsonDict | None):
        """
        This is a context manager that will reset the ChalkContext upon __exit__.
        :param context: If None, does nothing. Otherwise, sets the value of ChalkContext to the given dict.
        """
        token: Token | None = None
        try:
            if context is not None and len(context) > 0:
                token = cls._context_var.set(context)
            yield
        finally:
            if token is not None:
                cls._context_var.reset(token)

    @classmethod
    def get(cls, key: str, default: Optional[JsonValue] = None) -> Optional[JsonValue]:
        """

        Parameters
        ----------
        key
            The key to get from the context.
        default
            The default value to return if the key is not found. None by default.

        Returns
        -------
        str | int | float | bool | None
            The value associated with the key in the context, or the default value if the key is not found.
        """
        context_dict = cls._context_var.get()
        if context_dict is None:
            return default
        return context_dict.get(key, default)

    @classmethod
    def _reset(cls, token: Token | None):
        if token is None:
            return
        cls._context_var.reset(token)



from chalk._gen.chalk.auth.v1 import audit_pb2 as _audit_pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.server.v1 import link_pb2 as _link_pb2
from chalk._gen.chalk.utils.v1 import sensitive_pb2 as _sensitive_pb2
from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetTokenRequest(_message.Message):
    __slots__ = ("client_id", "client_secret", "grant_type", "scope")
    CLIENT_ID_FIELD_NUMBER: _ClassVar[int]
    CLIENT_SECRET_FIELD_NUMBER: _ClassVar[int]
    GRANT_TYPE_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    client_id: str
    client_secret: str
    grant_type: str
    scope: str
    def __init__(
        self,
        client_id: _Optional[str] = ...,
        client_secret: _Optional[str] = ...,
        grant_type: _Optional[str] = ...,
        scope: _Optional[str] = ...,
    ) -> None: ...

class GetTokenResponse(_message.Message):
    __slots__ = (
        "access_token",
        "token_type",
        "expires_in",
        "expires_at",
        "api_server",
        "primary_environment",
        "engines",
        "grpc_engines",
        "environment_id_to_name",
    )
    class EnginesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    class GrpcEnginesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    class EnvironmentIdToNameEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    ACCESS_TOKEN_FIELD_NUMBER: _ClassVar[int]
    TOKEN_TYPE_FIELD_NUMBER: _ClassVar[int]
    EXPIRES_IN_FIELD_NUMBER: _ClassVar[int]
    EXPIRES_AT_FIELD_NUMBER: _ClassVar[int]
    API_SERVER_FIELD_NUMBER: _ClassVar[int]
    PRIMARY_ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    ENGINES_FIELD_NUMBER: _ClassVar[int]
    GRPC_ENGINES_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_TO_NAME_FIELD_NUMBER: _ClassVar[int]
    access_token: str
    token_type: str
    expires_in: int
    expires_at: _timestamp_pb2.Timestamp
    api_server: str
    primary_environment: str
    engines: _containers.ScalarMap[str, str]
    grpc_engines: _containers.ScalarMap[str, str]
    environment_id_to_name: _containers.ScalarMap[str, str]
    def __init__(
        self,
        access_token: _Optional[str] = ...,
        token_type: _Optional[str] = ...,
        expires_in: _Optional[int] = ...,
        expires_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        api_server: _Optional[str] = ...,
        primary_environment: _Optional[str] = ...,
        engines: _Optional[_Mapping[str, str]] = ...,
        grpc_engines: _Optional[_Mapping[str, str]] = ...,
        environment_id_to_name: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class UpdateLinkSessionRequest(_message.Message):
    __slots__ = ("status", "user_id", "session_id")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    USER_ID_FIELD_NUMBER: _ClassVar[int]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    status: str
    user_id: str
    session_id: str
    def __init__(
        self, status: _Optional[str] = ..., user_id: _Optional[str] = ..., session_id: _Optional[str] = ...
    ) -> None: ...

class UpdateLinkSessionResponse(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...



from chalk._gen.chalk.aggregate.v1 import backfill_pb2 as _backfill_pb2
from chalk._gen.chalk.aggregate.v1 import timeseries_pb2 as _timeseries_pb2
from chalk._gen.chalk.auth.v1 import audit_pb2 as _audit_pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class PlanAggregateBackfillRequest(_message.Message):
    __slots__ = ("params",)
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    params: _backfill_pb2.AggregateBackfillUserParams
    def __init__(
        self, params: _Optional[_Union[_backfill_pb2.AggregateBackfillUserParams, _Mapping]] = ...
    ) -> None: ...

class PlanAggregateBackfillResponse(_message.Message):
    __slots__ = ("estimate", "errors", "backfills", "aggregate_backfill_id")
    ESTIMATE_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    BACKFILLS_FIELD_NUMBER: _ClassVar[int]
    AGGREGATE_BACKFILL_ID_FIELD_NUMBER: _ClassVar[int]
    estimate: _backfill_pb2.AggregateBackfillCostEstimate
    errors: _containers.RepeatedScalarFieldContainer[str]
    backfills: _containers.RepeatedCompositeFieldContainer[_backfill_pb2.AggregateBackfillWithCostEstimate]
    aggregate_backfill_id: str
    def __init__(
        self,
        estimate: _Optional[_Union[_backfill_pb2.AggregateBackfillCostEstimate, _Mapping]] = ...,
        errors: _Optional[_Iterable[str]] = ...,
        backfills: _Optional[_Iterable[_Union[_backfill_pb2.AggregateBackfillWithCostEstimate, _Mapping]]] = ...,
        aggregate_backfill_id: _Optional[str] = ...,
    ) -> None: ...

class GetAggregatesRequest(_message.Message):
    __slots__ = ("for_features",)
    FOR_FEATURES_FIELD_NUMBER: _ClassVar[int]
    for_features: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, for_features: _Optional[_Iterable[str]] = ...) -> None: ...

class GetAggregatesResponse(_message.Message):
    __slots__ = ("series", "errors")
    SERIES_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    series: _containers.RepeatedCompositeFieldContainer[_timeseries_pb2.AggregateTimeSeries]
    errors: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        series: _Optional[_Iterable[_Union[_timeseries_pb2.AggregateTimeSeries, _Mapping]]] = ...,
        errors: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class GetAggregateBackfillJobsRequest(_message.Message):
    __slots__ = ("limit", "cursor", "plan_hash")
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    CURSOR_FIELD_NUMBER: _ClassVar[int]
    PLAN_HASH_FIELD_NUMBER: _ClassVar[int]
    limit: int
    cursor: str
    plan_hash: str
    def __init__(
        self, limit: _Optional[int] = ..., cursor: _Optional[str] = ..., plan_hash: _Optional[str] = ...
    ) -> None: ...

class GetAggregateBackfillJobsResponse(_message.Message):
    __slots__ = ("jobs", "cursor")
    JOBS_FIELD_NUMBER: _ClassVar[int]
    CURSOR_FIELD_NUMBER: _ClassVar[int]
    jobs: _containers.RepeatedCompositeFieldContainer[_backfill_pb2.AggregateBackfillJob]
    cursor: str
    def __init__(
        self,
        jobs: _Optional[_Iterable[_Union[_backfill_pb2.AggregateBackfillJob, _Mapping]]] = ...,
        cursor: _Optional[str] = ...,
    ) -> None: ...

class GetAggregateBackfillJobRequest(_message.Message):
    __slots__ = ("aggregate_backfill_id",)
    AGGREGATE_BACKFILL_ID_FIELD_NUMBER: _ClassVar[int]
    aggregate_backfill_id: str
    def __init__(self, aggregate_backfill_id: _Optional[str] = ...) -> None: ...

class GetAggregateBackfillJobResponse(_message.Message):
    __slots__ = ("job",)
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: _backfill_pb2.AggregateBackfillJob
    def __init__(self, job: _Optional[_Union[_backfill_pb2.AggregateBackfillJob, _Mapping]] = ...) -> None: ...

class GetCronAggregateBackfillRequest(_message.Message):
    __slots__ = ("cron_aggregate_backfill_id", "plan_hash")
    CRON_AGGREGATE_BACKFILL_ID_FIELD_NUMBER: _ClassVar[int]
    PLAN_HASH_FIELD_NUMBER: _ClassVar[int]
    cron_aggregate_backfill_id: str
    plan_hash: str
    def __init__(self, cron_aggregate_backfill_id: _Optional[str] = ..., plan_hash: _Optional[str] = ...) -> None: ...

class GetCronAggregateBackfillResponse(_message.Message):
    __slots__ = ("cron_aggregate_backfill",)
    CRON_AGGREGATE_BACKFILL_FIELD_NUMBER: _ClassVar[int]
    cron_aggregate_backfill: _backfill_pb2.CronAggregateBackfill
    def __init__(
        self, cron_aggregate_backfill: _Optional[_Union[_backfill_pb2.CronAggregateBackfill, _Mapping]] = ...
    ) -> None: ...

class GetActiveCronAggregateBackfillsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class CronAggregateBackfillWithLatestRun(_message.Message):
    __slots__ = ("cron_aggregate_backfill", "latest_job")
    CRON_AGGREGATE_BACKFILL_FIELD_NUMBER: _ClassVar[int]
    LATEST_JOB_FIELD_NUMBER: _ClassVar[int]
    cron_aggregate_backfill: _backfill_pb2.CronAggregateBackfill
    latest_job: _backfill_pb2.AggregateBackfillJob
    def __init__(
        self,
        cron_aggregate_backfill: _Optional[_Union[_backfill_pb2.CronAggregateBackfill, _Mapping]] = ...,
        latest_job: _Optional[_Union[_backfill_pb2.AggregateBackfillJob, _Mapping]] = ...,
    ) -> None: ...

class GetActiveCronAggregateBackfillsResponse(_message.Message):
    __slots__ = ("cron_aggregate_backfills",)
    CRON_AGGREGATE_BACKFILLS_FIELD_NUMBER: _ClassVar[int]
    cron_aggregate_backfills: _containers.RepeatedCompositeFieldContainer[CronAggregateBackfillWithLatestRun]
    def __init__(
        self, cron_aggregate_backfills: _Optional[_Iterable[_Union[CronAggregateBackfillWithLatestRun, _Mapping]]] = ...
    ) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/timeserieschart.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n%chalk/server/v1/timeserieschart.proto\x12\x0f\x63halk.server.v1\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto"2\n\x07PointV2\x12\x19\n\x05value\x18\x01 \x01(\x01H\x00R\x05value\x88\x01\x01:\x02\x18\x01\x42\x08\n\x06_value"p\n\x0cTimeSeriesV2\x12\x30\n\x06points\x18\x01 \x03(\x0b\x32\x18.chalk.server.v1.PointV2R\x06points\x12\x14\n\x05label\x18\x02 \x01(\tR\x05label\x12\x14\n\x05units\x18\x03 \x01(\tR\x05units:\x02\x18\x01"\xdb\x01\n\x11TimeSeriesChartV2\x12\x14\n\x05title\x18\x01 \x01(\tR\x05title\x12\x35\n\x06series\x18\x02 \x03(\x0b\x32\x1d.chalk.server.v1.TimeSeriesV2R\x06series\x12\x35\n\x08x_series\x18\x03 \x03(\x0b\x32\x1a.google.protobuf.TimestampR\x07xSeries\x12>\n\rwindow_period\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0cwindowPeriod:\x02\x18\x01\x42\x9d\x01\n\x13\x63om.chalk.server.v1B\x14TimeserieschartProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.timeserieschart_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\024TimeserieschartProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_POINTV2"]._options = None
    _globals["_POINTV2"]._serialized_options = b"\030\001"
    _globals["_TIMESERIESV2"]._options = None
    _globals["_TIMESERIESV2"]._serialized_options = b"\030\001"
    _globals["_TIMESERIESCHARTV2"]._options = None
    _globals["_TIMESERIESCHARTV2"]._serialized_options = b"\030\001"
    _globals["_POINTV2"]._serialized_start = 123
    _globals["_POINTV2"]._serialized_end = 173
    _globals["_TIMESERIESV2"]._serialized_start = 175
    _globals["_TIMESERIESV2"]._serialized_end = 287
    _globals["_TIMESERIESCHARTV2"]._serialized_start = 290
    _globals["_TIMESERIESCHARTV2"]._serialized_end = 509
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/manager.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1d\x63halk/server/v1/manager.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto"B\n\x1dGetClusterEnvironmentsRequest\x12!\n\x0c\x63luster_name\x18\x01 \x01(\tR\x0b\x63lusterName"I\n\x1eGetClusterEnvironmentsResponse\x12\'\n\x0f\x65nvironment_ids\x18\x01 \x03(\tR\x0e\x65nvironmentIds2\x94\x01\n\x0eManagerService\x12\x81\x01\n\x16GetClusterEnvironments\x12..chalk.server.v1.GetClusterEnvironmentsRequest\x1a/.chalk.server.v1.GetClusterEnvironmentsResponse"\x06\x90\x02\x01\x80}\x02\x42\x95\x01\n\x13\x63om.chalk.server.v1B\x0cManagerProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.manager_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\014ManagerProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_MANAGERSERVICE"].methods_by_name["GetClusterEnvironments"]._options = None
    _globals["_MANAGERSERVICE"].methods_by_name["GetClusterEnvironments"]._serialized_options = b"\220\002\001\200}\002"
    _globals["_GETCLUSTERENVIRONMENTSREQUEST"]._serialized_start = 83
    _globals["_GETCLUSTERENVIRONMENTSREQUEST"]._serialized_end = 149
    _globals["_GETCLUSTERENVIRONMENTSRESPONSE"]._serialized_start = 151
    _globals["_GETCLUSTERENVIRONMENTSRESPONSE"]._serialized_end = 224
    _globals["_MANAGERSERVICE"]._serialized_start = 227
    _globals["_MANAGERSERVICE"]._serialized_end = 375
# @@protoc_insertion_point(module_scope)



from typing import Union, cast

import pyarrow as pa

from chalk._gen.chalk.arrow.v1 import arrow_pb2 as arrow_pb
from chalk._gen.chalk.expression.v1 import expression_pb2 as expr_pb
from chalk.features._encoding.converter import PrimitiveFeatureConverter
from chalk.features._encoding.primitive import TPrimitive


def convert_pa_dtype_to_proto_expr(dtype: pa.DataType) -> expr_pb.LogicalExprNode:
    """
    This is kind of a hack - use the 'null' literal for a dtype to convey the dtype information.
    """

    return expr_pb.LogicalExprNode(
        literal_value=expr_pb.ExprLiteral(
            # HACK: Using this to store a dtype. This is not really a scalar value.
            value=arrow_pb.ScalarValue(null_value=PrimitiveFeatureConverter.convert_pa_dtype_to_proto_dtype(dtype)),
            is_arrow_scalar_object=True,
        )
    )


def convert_literal_to_proto_expr(value: Union[TPrimitive, pa.DataType]) -> expr_pb.LogicalExprNode:
    is_arrow_scalar_object = False
    if isinstance(value, pa.Scalar):
        pa_dtype = value.type  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
        is_arrow_scalar_object = True
    elif isinstance(value, pa.DataType):
        return convert_pa_dtype_to_proto_expr(value)
    else:
        try:
            pa_dtype = pa.scalar(value).type
        except Exception as e:
            raise ValueError(f"Could not infer literal type for value `{value}`") from e
    converter = PrimitiveFeatureConverter(
        name="convert_literal_to_proto_expr",
        is_nullable=False,
        pyarrow_dtype=pa_dtype,
    )
    return expr_pb.LogicalExprNode(
        literal_value=expr_pb.ExprLiteral(
            value=converter.from_primitive_to_protobuf(value), is_arrow_scalar_object=is_arrow_scalar_object
        )
    )


def convert_proto_expr_to_literal(node: expr_pb.LogicalExprNode) -> TPrimitive:
    if not node.HasField("literal_value"):
        raise ValueError("Expected a literal expression")
    scalar_val = PrimitiveFeatureConverter.from_protobuf_to_pyarrow(node.literal_value.value)
    if node.literal_value.is_arrow_scalar_object:
        return scalar_val
    else:
        return cast(TPrimitive, scalar_val.as_py())



from __future__ import annotations

import collections.abc
import datetime
import enum
import functools
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Protocol, Sequence, TypeVar, Union, cast

import pyarrow as pa
from typing_extensions import Self, TypeGuard

from chalk.features._encoding.converter import pyarrow_to_polars
from chalk.features.feature_field import Feature
from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
from chalk.features.filter import Filter, TimeDelta, get_filter_now
from chalk.utils.collections import ensure_tuple

if TYPE_CHECKING:
    import polars as pl


class ParsedExpr(Protocol):
    def __eq__(self, __value: Any, /) -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        ...

    def __ne__(self, __value: Any, /) -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        ...

    def __gt__(self, __value: Any, /) -> Self:
        ...

    def __lt__(self, __value: Any, /) -> Self:
        ...

    def __le__(self, __value: Any, /) -> Self:
        ...

    def __ge__(self, __value: Any, /) -> Self:
        ...

    def __invert__(self) -> Self:
        ...

    def __add__(self, __value: Any, /) -> Self:
        ...

    def __mul__(self, __value: Any, /) -> Self:
        ...

    def __sub__(self, __value: Any, /) -> Self:
        ...

    def __and__(self, __value: Any, /) -> Self:
        ...

    def __or__(self, __value: Any, /) -> Self:
        ...


TExpr = TypeVar("TExpr", bound=ParsedExpr)


class FilterConverter(Protocol[TExpr]):
    """
    Converts a Filter expression tree into a polars or pyarrow expression
    that can be used to filter tables of the corresponding type.
    """

    def cast(self, expr: TExpr, dtype: pa.DataType) -> TExpr:
        ...

    def _is_expr(self, o: Any) -> TypeGuard[TExpr]:
        """Need this for isinstance() checks, TypeVar is insufficient :-("""
        ...

    def col(self, name: str) -> TExpr:
        """Expression for selecting a column by the given name"""
        ...

    def lit(self, value: Any, dtype: pa.DataType[Any]) -> TExpr:
        """Expression for a literal value"""
        ...

    def is_null(self, expr: TExpr) -> TExpr:
        """Expression for checking null"""
        ...

    def call(self, function_name: str, *args: TExpr) -> TExpr:
        """Call a system-defined function."""
        ...

    def is_not_null(self, expr: TExpr) -> TExpr:
        """Should be equivalent to `not is_null(expr)`."""
        ...

    def if_else(self, cond: TExpr, true_case: TExpr, false_case: TExpr) -> TExpr:
        """For every row where ``cond`` is true, return the ``true_case``. Otherwise, return ``false_case``."""
        ...

    def fill_null(self, expr: TExpr, fill_with: TExpr) -> TExpr:
        return self.if_else(self.is_null(expr), fill_with, expr)

    def _expr_is_in_sequence_lit(self, expr: TExpr, values: Sequence[TExpr], dtype: pa.DataType) -> TExpr:
        """
        Returns an expression that checks whether `expr` is in a sequence of literal values.
        i.e. analogous to `lambda x: expr(x) in values`
        """
        # Using expr == values[0] OR expr == values[2] OR expr == values[3] so we can do our fuzzy eq_missing logic
        exprs = [self._eq_missing(expr, x, dtype) for x in values]
        if len(exprs) == 0:
            return self.lit(False, pa.bool_())
        return functools.reduce(lambda a, b: a | b, exprs)

    def convert_filters(
        self,
        filters: Sequence[Filter],
        df_schema: Mapping[str, pa.DataType],
        timestamp_feature: Optional[Feature] = None,
        now: Optional[Union[datetime.datetime, TExpr]] = None,
        now_col_name: Optional[str] = None,
    ) -> Optional[TExpr]:
        if len(filters) == 0:
            return None
        parsed_filters = [
            self._convert_filter_to_expr(f, df_schema, timestamp_feature, now, now_col_name) for f in filters
        ]
        return functools.reduce(lambda a, b: a & b, parsed_filters)

    def _eq_missing(self, lhs: TExpr, rhs: TExpr, dtype: pa.DataType) -> TExpr:
        # If either side is a struct, then recursively do eq_missing on each struct member
        if pa.types.is_struct(dtype):
            assert isinstance(dtype, pa.StructType)
            sub_exprs: List[TExpr] = []
            sa = self._get_struct_adapter(dtype)
            for f in dtype:
                lhs_expr, lhs_dtype = sa.get_expr_for_field(lhs, f.name)
                rhs_expr, rhs_dtype = sa.get_expr_for_field(rhs, f.name)
                if lhs_dtype != rhs_dtype:
                    raise TypeError(
                        f"Struct field '{f.name}' has a different lhs dtype ({lhs_dtype}) than the RHS dtype ({rhs_dtype})"
                    )
                sub_exprs.append(self._eq_missing(lhs_expr, rhs_expr, lhs_dtype))
            if len(sub_exprs) == 0:
                sub_ans = self.lit(True, pa.bool_())
            else:
                sub_ans = functools.reduce(lambda a, b: a & b, sub_exprs)
        else:
            # FIXME: It is too hard to do eq_missing for sub elements of variable size lists. Probably need to do some sort of join / aggregation
            sub_ans = lhs == rhs
        return self.if_else(
            self.is_null(lhs) == self.is_null(rhs),
            self.fill_null(
                # If both sides have the same nullability, return the sub answer, coalescing null values to true (b/c we want null==null)
                sub_ans,
                self.lit(True, pa.bool_()),
            ),
            self.lit(
                # If both sides have different nullability, then return False, because Null != (anything other than null)
                False,
                pa.bool_(),
            ),
        )

    def _convert_filter_to_expr(
        self,
        f: Filter,
        df_schema: Mapping[str, pa.DataType],
        timestamp_feature: Optional[Feature] = None,
        now: Optional[Union[datetime.datetime, TExpr]] = None,
        now_col_name: Optional[str] = None,
    ) -> TExpr:
        # Passing `now` in explicitly instead of using datetime.datetime.now() so that multiple filters
        # relying on relative timestamps (e.g. before, after) will have the same "now" time.
        def _convert_sub_filter(sub_filter: Filter, side_name: str):
            assert isinstance(sub_filter, Filter), f"{side_name} must be a filter"
            return self._convert_filter_to_expr(sub_filter, df_schema, timestamp_feature, now, now_col_name)

        if f.operation == "not" or f.operation == "~":
            assert isinstance(f.lhs, Filter)
            assert f.rhs is None, "not has just one side"
            return ~_convert_sub_filter(f.lhs, "lhs")
        elif f.operation == "and" or f.operation == "&":
            return _convert_sub_filter(f.lhs, "lhs") & _convert_sub_filter(f.rhs, "rhs")
        elif f.operation == "or" or f.operation == "|":
            return _convert_sub_filter(f.lhs, "lhs") | _convert_sub_filter(f.rhs, "rhs")

        lhs = self._parse_feature_or_value(f.lhs, timestamp_feature, now, now_col_name)
        rhs = self._parse_feature_or_value(f.rhs, timestamp_feature, now, now_col_name)

        lhs_converter = None
        dtype = None
        if isinstance(lhs, Feature):
            lhs_converter = lhs.converter
            col_name = str(lhs)
            key_error_or_none = dataframe_missing_key_error([col_name], tuple(df_schema.keys()))
            if key_error_or_none is not None:
                raise key_error_or_none
            lhs = self.col(col_name)
            dtype = lhs_converter.pyarrow_dtype

        rhs_converter = None
        if isinstance(rhs, Feature):
            rhs_converter = rhs.converter
            col_name = str(rhs)
            key_error_or_none = dataframe_missing_key_error([col_name], tuple(df_schema.keys()))
            if key_error_or_none is not None:
                raise key_error_or_none
            rhs = self.col(col_name)
            dtype = rhs_converter.pyarrow_dtype
        assert dtype is not None, "One side must be an expression"

        if lhs_converter is None:
            # LHS is literal. Encode it into the rhs_dtype
            assert rhs_converter is not None, "One side of the filter must be an expression"
            if not self._is_expr(lhs):
                lhs = self.lit(rhs_converter.from_rich_to_primitive(lhs), rhs_converter.pyarrow_dtype)
        if rhs_converter is None:
            # RHS is literal. Encode it into the lhs_dtype
            assert lhs_converter is not None, "One side of the filter must be an expression"
            if not self._is_expr(rhs):
                if f.operation in ("in", "not in"):
                    assert isinstance(rhs, collections.abc.Iterable)
                    rhs = [self.lit(lhs_converter.from_rich_to_primitive(x), lhs_converter.pyarrow_dtype) for x in rhs]
                else:
                    rhs = self.lit(lhs_converter.from_rich_to_primitive(rhs), lhs_converter.pyarrow_dtype)

        if f.operation in ("in", "not in"):
            assert lhs_converter is not None
            assert self._is_expr(lhs)
            ret = self._expr_is_in_sequence_lit(lhs, cast(Any, rhs), lhs_converter.pyarrow_dtype)
            if f.operation == "not in":
                ret = ~ret
        elif f.operation == "==" or f.operation == "!=":
            if lhs_converter is None or rhs_converter is None:
                # If comparing against a literal value, then use eq_missing
                assert self._is_expr(rhs), "The rhs must be an expression if doing equality"
                ret = self._eq_missing(lhs, rhs, dtype)
            else:
                # If both are not None, then do a normal eq
                ret = lhs == rhs
            if f.operation == "!=":
                # Invert it
                ret = ~ret
        elif f.operation == ">=":
            ret = lhs >= rhs
        elif f.operation == ">":
            ret = lhs > rhs
        elif f.operation == "<":
            ret = lhs < rhs
        elif f.operation == "<=":
            ret = lhs <= rhs
        else:
            raise ValueError(f'Unknown operation "{f.operation}"')
        assert self._is_expr(ret)
        return ret

    def _parse_feature_or_value(
        self,
        f: Union[Feature, Any],
        timestamp_feature: Optional[Feature],
        now: Optional[Union[datetime.datetime, TExpr]],
        now_column_name: Optional[str],
    ):
        """Parse a feature or value into the correct type that can be used for filtering."""
        f = self._feature_type_or_value(f)
        f = self._maybe_convert_timedelta_to_timestamp(f, now, now_column_name)
        f = self._maybe_replace_timestamp_feature(f, timestamp_feature)
        if isinstance(f, enum.Enum):
            f = f.value
        return f

    def _feature_type_or_value(self, e: Union[Feature, FeatureWrapper, Any]) -> Union[Feature, Any]:
        if isinstance(e, FeatureWrapper):
            e = unwrap_feature(e)
        return e

    def _maybe_convert_timedelta_to_timestamp(
        self,
        f: Union[TimeDelta, datetime.timedelta, Any],
        now: Optional[Union[datetime.datetime, TExpr]],
        now_column_name: Optional[str] = None,
    ) -> Union[TExpr, datetime.datetime, Any]:
        """Convert timedeltas relative to ``now`` into absolute datetimes."""
        if now is not None and now_column_name is not None:
            raise ValueError(
                (
                    "Can't specify both now and now_column_name -- one or the other must be used as a point of reference for "
                    "the time delta"
                )
            )
        if isinstance(f, TimeDelta):
            f = f.to_std()
        if isinstance(f, datetime.timedelta):
            if now is None and now_column_name is None:
                raise ValueError(
                    (
                        "The filter contains a relative timestamp. The current datetime or current date column "
                        "must be provided to evaluate this filter."
                    )
                )
            if now_column_name is not None:
                # creates a polars/pyarrow expression∂ that evaluates to dates that are within the timedelta specified by `f`
                return self.col(now_column_name) + self.lit(f, pa.duration("us"))
            assert now is not None
            if not self._is_expr(now):
                now = self.lit(now, pa.timestamp("us", "UTC"))
            return now + self.lit(f, pa.duration("us"))

        return f

    def _maybe_replace_timestamp_feature(
        self, f: Union[Feature, Any], observed_at_feature: Optional[Feature]
    ) -> Feature:
        """Replace the ``CHALK_TS`` pseudo-feature with the actual timestamp column."""
        if not isinstance(f, Feature) or f.fqn != "__chalk__.CHALK_TS":
            return f
        if observed_at_feature is not None:
            return observed_at_feature
        raise ValueError("No Timestamp Feature Found")

    def _get_struct_adapter(self, dtype: pa.StructType) -> StructAdapter[TExpr]:
        ...


class PolarsFilterConverter(FilterConverter["pl.Expr"]):
    def _is_expr(self, o: Any) -> TypeGuard[pl.Expr]:
        import polars as pl

        return isinstance(o, pl.Expr)

    def cast(self, expr: pl.Expr, dtype: pa.DataType) -> pl.Expr:
        raise NotImplementedError

    def col(self, name: str) -> pl.Expr:
        import polars as pl

        return pl.col(name)

    def lit(self, value: Any, dtype: pa.DataType) -> pl.Expr:
        import polars as pl

        if pa.types.is_struct(dtype):
            # Polars cannot handle structs directly.
            # Instead, we will convert each field, and wrap it in a pl.struct
            assert isinstance(dtype, pa.StructType)
            sub_fields = []
            for f in dtype:
                name = f.name
                try:
                    val = value[name]
                except KeyError:
                    val = getattr(value, name)
                except AttributeError:
                    raise ValueError(f"Literal value has no attribute or key for field '{name}'")
                sub_fields.append(self.lit(val, f.type).alias(name))
            return pl.struct(sub_fields)

        if pa.types.is_list(dtype) or pa.types.is_large_list(dtype) or pa.types.is_fixed_size_list(dtype):
            # If it's a list, then we will convert each value by hand, and return a polars literal of these values
            assert isinstance(dtype, (pa.ListType, pa.LargeListType, pa.FixedSizeListType))
            return pl.lit([self.lit(x, dtype.value_type) for x in value])

        pl_dtype = pyarrow_to_polars(dtype)

        if (
            isinstance(value, datetime.datetime)
            and value.tzinfo is not None
            and isinstance(pl_dtype, pl.Datetime)
            and pl_dtype.time_zone is not None
        ):
            # If given a datetype with a timezone, and we have a target timezone on the datetime type,
            # we will first interpret the datetime in polars with the provided time zone, and then convert it to the desired time zone
            # Polars does not allow literal datetimes with a timezone to be directly interpreted as a datetime with time zone
            assert pl_dtype.time_unit is not None, f"polars dtype {pl_dtype} incorrectly set with no time unit."
            return pl.lit(value, pl.Datetime(pl_dtype.time_unit)).dt.convert_time_zone(pl_dtype.time_zone)

        return pl.lit(value, dtype=pl_dtype)

    def is_null(self, expr: pl.Expr) -> pl.Expr:
        return expr.is_null()

    def is_not_null(self, expr: pl.Expr) -> pl.Expr:
        return expr.is_not_null()

    def if_else(self, cond: pl.Expr, true_case: pl.Expr, false_case: pl.Expr) -> pl.Expr:
        import polars as pl

        return pl.when(cond).then(true_case).otherwise(false_case)

    def _get_struct_adapter(self, dtype: pa.StructType) -> StructAdapter[pl.Expr]:
        return _PolarsStructAdapter(dtype)

    def call(self, function_name: str, *args: pl.Expr) -> pl.Expr:
        import polars as pl

        well_known_functions = {
            "greatest": pl.max_horizontal,
            "least": pl.min_horizontal,
            "coalesce": pl.coalesce,
        }
        try:
            fn = well_known_functions[function_name]
        except KeyError:
            raise ValueError(f"Unrecognized function {function_name}.")
        return fn(*args)


def convert_filters_to_pl_expr(
    filters: Sequence[Filter],
    df_schema: Mapping[str, pl.PolarsDataType],
    timestamp_feature: Optional[Feature] = None,
    now: Optional[datetime.datetime] = None,
    now_col_name: Optional[str] = None,
) -> Optional[pl.Expr]:
    import polars as pl

    tbl = pl.DataFrame({k: [] for k in df_schema}, df_schema).to_arrow()
    pa_schema = {k: v for (k, v) in zip(tbl.schema.names, tbl.schema.types)}
    ans = PolarsFilterConverter().convert_filters(filters, pa_schema, timestamp_feature, now, now_col_name)
    assert isinstance(ans, pl.Expr) or ans is None
    return ans


class StructAdapter(Protocol[TExpr]):
    """
    Wrapper for a polars or pyarrow struct.
    """

    def get_expr_for_field(self, expr: TExpr, name: str) -> tuple[TExpr, pa.DataType]:
        ...


class _PolarsStructAdapter(StructAdapter["pl.Expr"]):
    def __init__(self, struct_type: pa.StructType):
        super().__init__()
        self._struct_type = struct_type
        self._name_to_dtype = {f.name: f.type for f in struct_type}

    def get_expr_for_field(self, expr: pl.Expr, name: str) -> tuple[pl.Expr, pa.DataType]:
        return expr.struct.field(name), self._name_to_dtype[name]


def filter_data_frame(
    item: Any,
    underlying: Union[pl.DataFrame, pl.LazyFrame],
    namespace: Optional[str],
) -> Union[pl.DataFrame, pl.LazyFrame]:

    # Use the Chalk projection / selection syntax, where we support our Filter objects and
    # selection by column name

    projections: list[str] = []
    filters: List[Filter] = []
    for x in ensure_tuple(item):
        if isinstance(x, (FeatureWrapper, Feature, str)):
            projections.append(str(x))

        elif isinstance(x, Filter):
            filters.append(x)
        else:
            raise TypeError(
                "When indexing by Filters or Features, it is not simultaneously possible to perform other indexing operations."
            )
    now = get_filter_now()
    if len(projections) > 0:
        key_error_or_none = dataframe_missing_key_error(projections, underlying.columns)
        if key_error_or_none is not None:
            raise key_error_or_none
    # now = datetime.datetime.now(tz=datetime.timezone.utc)
    from chalk.features.feature_set import CURRENT_FEATURE_REGISTRY

    timestamp_feature = (
        None if namespace is None else CURRENT_FEATURE_REGISTRY.get().get_feature_sets()[namespace].__chalk_ts__
    )
    pl_expr = convert_filters_to_pl_expr(filters, underlying.schema, timestamp_feature, now)
    df = underlying
    if pl_expr is not None:
        df = df.filter(pl_expr)
    # Do the projection
    if len(projections) > 0:
        df = df.select(projections)
    return df


def dataframe_missing_key_error(prospective_columns: Sequence[str], schema: Sequence[str]) -> Union[None, KeyError]:
    cols_set = set(schema)
    missing_cols = [c for c in prospective_columns if c not in cols_set]
    if len(missing_cols) > 0:
        return KeyError(
            (
                f"Attempted to select missing columns [{', '.join(sorted(missing_cols))}] "
                f"from DataFrame with columns [{', '.join(sorted(list(cols_set)))}]"
            )
        )
    return None



from chalk._gen.chalk.common.v1 import chalk_error_pb2 as _chalk_error_pb2
from chalk._gen.chalk.common.v1 import query_status_pb2 as _query_status_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class DatasetSampleFilter(_message.Message):
    __slots__ = ("lower_bound", "upper_bound", "max_samples")
    LOWER_BOUND_FIELD_NUMBER: _ClassVar[int]
    UPPER_BOUND_FIELD_NUMBER: _ClassVar[int]
    MAX_SAMPLES_FIELD_NUMBER: _ClassVar[int]
    lower_bound: _timestamp_pb2.Timestamp
    upper_bound: _timestamp_pb2.Timestamp
    max_samples: int
    def __init__(
        self,
        lower_bound: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        upper_bound: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        max_samples: _Optional[int] = ...,
    ) -> None: ...

class DatasetFilter(_message.Message):
    __slots__ = ("sample_filters", "max_cache_age_secs")
    SAMPLE_FILTERS_FIELD_NUMBER: _ClassVar[int]
    MAX_CACHE_AGE_SECS_FIELD_NUMBER: _ClassVar[int]
    sample_filters: DatasetSampleFilter
    max_cache_age_secs: float
    def __init__(
        self,
        sample_filters: _Optional[_Union[DatasetSampleFilter, _Mapping]] = ...,
        max_cache_age_secs: _Optional[float] = ...,
    ) -> None: ...

class DatasetRevisionResponse(_message.Message):
    __slots__ = (
        "dataset_name",
        "dataset_id",
        "environment_id",
        "revision_id",
        "creator_id",
        "outputs",
        "givens_uri",
        "status",
        "filters",
        "num_partitions",
        "num_bytes",
        "output_uris",
        "output_version",
        "branch",
        "dashboard_url",
        "created_at",
        "started_at",
        "terminated_at",
    )
    DATASET_NAME_FIELD_NUMBER: _ClassVar[int]
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    REVISION_ID_FIELD_NUMBER: _ClassVar[int]
    CREATOR_ID_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    GIVENS_URI_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    NUM_PARTITIONS_FIELD_NUMBER: _ClassVar[int]
    NUM_BYTES_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_URIS_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_VERSION_FIELD_NUMBER: _ClassVar[int]
    BRANCH_FIELD_NUMBER: _ClassVar[int]
    DASHBOARD_URL_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    STARTED_AT_FIELD_NUMBER: _ClassVar[int]
    TERMINATED_AT_FIELD_NUMBER: _ClassVar[int]
    dataset_name: str
    dataset_id: str
    environment_id: str
    revision_id: str
    creator_id: str
    outputs: _containers.RepeatedScalarFieldContainer[str]
    givens_uri: str
    status: _query_status_pb2.QueryStatus
    filters: DatasetFilter
    num_partitions: int
    num_bytes: int
    output_uris: str
    output_version: int
    branch: str
    dashboard_url: str
    created_at: _timestamp_pb2.Timestamp
    started_at: _timestamp_pb2.Timestamp
    terminated_at: _timestamp_pb2.Timestamp
    def __init__(
        self,
        dataset_name: _Optional[str] = ...,
        dataset_id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        revision_id: _Optional[str] = ...,
        creator_id: _Optional[str] = ...,
        outputs: _Optional[_Iterable[str]] = ...,
        givens_uri: _Optional[str] = ...,
        status: _Optional[_Union[_query_status_pb2.QueryStatus, str]] = ...,
        filters: _Optional[_Union[DatasetFilter, _Mapping]] = ...,
        num_partitions: _Optional[int] = ...,
        num_bytes: _Optional[int] = ...,
        output_uris: _Optional[str] = ...,
        output_version: _Optional[int] = ...,
        branch: _Optional[str] = ...,
        dashboard_url: _Optional[str] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        started_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        terminated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class DatasetResponse(_message.Message):
    __slots__ = ("is_finished", "version", "environment_id", "dataset_id", "dataset_name", "errors", "revisions")
    IS_FINISHED_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    DATASET_NAME_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    REVISIONS_FIELD_NUMBER: _ClassVar[int]
    is_finished: bool
    version: int
    environment_id: str
    dataset_id: str
    dataset_name: str
    errors: _containers.RepeatedCompositeFieldContainer[_chalk_error_pb2.ChalkError]
    revisions: _containers.RepeatedCompositeFieldContainer[DatasetRevisionResponse]
    def __init__(
        self,
        is_finished: bool = ...,
        version: _Optional[int] = ...,
        environment_id: _Optional[str] = ...,
        dataset_id: _Optional[str] = ...,
        dataset_name: _Optional[str] = ...,
        errors: _Optional[_Iterable[_Union[_chalk_error_pb2.ChalkError, _Mapping]]] = ...,
        revisions: _Optional[_Iterable[_Union[DatasetRevisionResponse, _Mapping]]] = ...,
    ) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/engine/v1/plan.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.arrow.v1 import arrow_pb2 as chalk_dot_arrow_dot_v1_dot_arrow__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1a\x63halk/engine/v1/plan.proto\x12\x0f\x63halk.engine.v1\x1a\x1a\x63halk/arrow/v1/arrow.proto""\n\x0cRawColumnKey\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name"-\n\x10\x46\x65\x61tureColumnKey\x12\x19\n\x08root_fqn\x18\x01 \x01(\tR\x07rootFqn"Z\n\x11HasManyFeatureKey\x12\x19\n\x08root_fqn\x18\x01 \x01(\tR\x07rootFqn\x12*\n\x02\x64\x66\x18\x02 \x01(\x0b\x32\x1a.chalk.engine.v1.DataFrameR\x02\x64\x66"h\n\x13\x44\x61taFrameFeatureKey\x12%\n\x0eroot_namespace\x18\x01 \x01(\tR\rrootNamespace\x12*\n\x02\x64\x66\x18\x02 \x01(\x0b\x32\x1a.chalk.engine.v1.DataFrameR\x02\x64\x66"\xcc\x01\n\tDataFrame\x12L\n\x10optional_columns\x18\x01 \x03(\x0b\x32!.chalk.engine.v1.FeatureColumnKeyR\x0foptionalColumns\x12L\n\x10required_columns\x18\x02 \x03(\x0b\x32!.chalk.engine.v1.FeatureColumnKeyR\x0frequiredColumns\x12\x19\n\x05limit\x18\x03 \x01(\x04H\x00R\x05limit\x88\x01\x01\x42\x08\n\x06_limit"\x84\x01\n\tColumnKey\x12=\n\x07\x66\x65\x61ture\x18\x01 \x01(\x0b\x32!.chalk.engine.v1.FeatureColumnKeyH\x00R\x07\x66\x65\x61ture\x12\x31\n\x03raw\x18\x02 \x01(\x0b\x32\x1d.chalk.engine.v1.RawColumnKeyH\x00R\x03rawB\x05\n\x03key"\x99\x01\n\x08TableKey\x12?\n\x08has_many\x18\x01 \x01(\x0b\x32".chalk.engine.v1.HasManyFeatureKeyH\x00R\x07hasMany\x12\x45\n\ndata_frame\x18\x02 \x01(\x0b\x32$.chalk.engine.v1.DataFrameFeatureKeyH\x00R\tdataFrameB\x05\n\x03key"\x91\x02\n\x03Key\x12>\n\nraw_column\x18\x01 \x01(\x0b\x32\x1d.chalk.engine.v1.RawColumnKeyH\x00R\trawColumn\x12;\n\x06scalar\x18\x02 \x01(\x0b\x32!.chalk.engine.v1.FeatureColumnKeyH\x00R\x06scalar\x12?\n\x08has_many\x18\x03 \x01(\x0b\x32".chalk.engine.v1.HasManyFeatureKeyH\x00R\x07hasMany\x12\x44\n\tdataframe\x18\x04 \x01(\x0b\x32$.chalk.engine.v1.DataFrameFeatureKeyH\x00R\tdataframeB\x06\n\x04impl"\xee\x03\n\rPyArrowSchema\x12\x45\n\x07scalars\x18\x01 \x03(\x0b\x32+.chalk.engine.v1.PyArrowSchema.ColumnSchemaR\x07scalars\x12\x42\n\x06groups\x18\x02 \x03(\x0b\x32*.chalk.engine.v1.PyArrowSchema.TableSchemaR\x06groups\x1a\xe0\x01\n\x0bTableSchema\x12+\n\x03key\x18\x01 \x01(\x0b\x32\x19.chalk.engine.v1.TableKeyR\x03key\x12N\n\x06schema\x18\x02 \x03(\x0b\x32\x36.chalk.engine.v1.PyArrowSchema.TableSchema.SchemaEntryR\x06schema\x1aT\n\x0bSchemaEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12/\n\x05value\x18\x02 \x01(\x0b\x32\x19.chalk.arrow.v1.ArrowTypeR\x05value:\x02\x38\x01\x1ao\n\x0c\x43olumnSchema\x12,\n\x03key\x18\x01 \x01(\x0b\x32\x1a.chalk.engine.v1.ColumnKeyR\x03key\x12\x31\n\x06schema\x18\x02 \x01(\x0b\x32\x19.chalk.arrow.v1.ArrowTypeR\x06schema"\x9e\x01\n\x04Plan\x12+\n\x05nodes\x18\x01 \x03(\x0b\x32\x15.chalk.engine.v1.NodeR\x05nodes\x12"\n\rroot_node_idx\x18\x02 \x01(\rR\x0brootNodeIdx\x12\x45\n\x0epyarrow_schema\x18\x03 \x01(\x0b\x32\x1e.chalk.engine.v1.PyArrowSchemaR\rpyarrowSchema"`\n\x04Node\x12)\n\x10\x63hildren_indices\x18\x01 \x03(\rR\x0f\x63hildrenIndices\x12-\n\x04impl\x18\x02 \x01(\x0b\x32\x19.chalk.engine.v1.NodeImplR\x04impl"\xd7\x02\n\x08NodeImpl\x12\x38\n\x07unknown\x18\x01 \x01(\x0b\x32\x1c.chalk.engine.v1.UnknownNodeH\x00R\x07unknown\x12>\n\x0bgivens_scan\x18\x02 \x01(\x0b\x32\x1b.chalk.engine.v1.GivensScanH\x00R\ngivensScan\x12\x34\n\x07project\x18\x03 \x01(\x0b\x32\x18.chalk.engine.v1.ProjectH\x00R\x07project\x12\x44\n\rchalk_project\x18\x04 \x01(\x0b\x32\x1d.chalk.engine.v1.ChalkProjectH\x00R\x0c\x63halkProject\x12M\n\x10\x64\x65\x66\x61ult_injector\x18\x05 \x01(\x0b\x32 .chalk.engine.v1.DefaultInjectorH\x00R\x0f\x64\x65\x66\x61ultInjectorB\x06\n\x04impl"*\n\x0bUnknownNode\x12\x1b\n\ttype_name\x18\x01 \x01(\tR\x08typeName":\n\nGivensScan\x12,\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x14.chalk.engine.v1.KeyR\x06\x66ields"x\n\x0c\x43halkProject\x12,\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x14.chalk.engine.v1.KeyR\x06\x66ields\x12:\n\x1apromote_ts_to_feature_time\x18\x02 \x01(\x08R\x16promoteTsToFeatureTime"7\n\x07Project\x12,\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x14.chalk.engine.v1.KeyR\x06\x66ields"X\n\x0f\x44\x65\x66\x61ultInjector\x12\x45\n\x13\x64\x65\x66\x61ulting_features\x18\x01 \x03(\x0b\x32\x14.chalk.engine.v1.KeyR\x12\x64\x65\x66\x61ultingFeaturesB~\n\x13\x63om.chalk.engine.v1B\tPlanProtoP\x01\xa2\x02\x03\x43\x45X\xaa\x02\x0f\x43halk.Engine.V1\xca\x02\x0f\x43halk\\Engine\\V1\xe2\x02\x1b\x43halk\\Engine\\V1\\GPBMetadata\xea\x02\x11\x43halk::Engine::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.engine.v1.plan_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.engine.v1B\tPlanProtoP\001\242\002\003CEX\252\002\017Chalk.Engine.V1\312\002\017Chalk\\Engine\\V1\342\002\033Chalk\\Engine\\V1\\GPBMetadata\352\002\021Chalk::Engine::V1"
    _globals["_PYARROWSCHEMA_TABLESCHEMA_SCHEMAENTRY"]._options = None
    _globals["_PYARROWSCHEMA_TABLESCHEMA_SCHEMAENTRY"]._serialized_options = b"8\001"
    _globals["_RAWCOLUMNKEY"]._serialized_start = 75
    _globals["_RAWCOLUMNKEY"]._serialized_end = 109
    _globals["_FEATURECOLUMNKEY"]._serialized_start = 111
    _globals["_FEATURECOLUMNKEY"]._serialized_end = 156
    _globals["_HASMANYFEATUREKEY"]._serialized_start = 158
    _globals["_HASMANYFEATUREKEY"]._serialized_end = 248
    _globals["_DATAFRAMEFEATUREKEY"]._serialized_start = 250
    _globals["_DATAFRAMEFEATUREKEY"]._serialized_end = 354
    _globals["_DATAFRAME"]._serialized_start = 357
    _globals["_DATAFRAME"]._serialized_end = 561
    _globals["_COLUMNKEY"]._serialized_start = 564
    _globals["_COLUMNKEY"]._serialized_end = 696
    _globals["_TABLEKEY"]._serialized_start = 699
    _globals["_TABLEKEY"]._serialized_end = 852
    _globals["_KEY"]._serialized_start = 855
    _globals["_KEY"]._serialized_end = 1128
    _globals["_PYARROWSCHEMA"]._serialized_start = 1131
    _globals["_PYARROWSCHEMA"]._serialized_end = 1625
    _globals["_PYARROWSCHEMA_TABLESCHEMA"]._serialized_start = 1288
    _globals["_PYARROWSCHEMA_TABLESCHEMA"]._serialized_end = 1512
    _globals["_PYARROWSCHEMA_TABLESCHEMA_SCHEMAENTRY"]._serialized_start = 1428
    _globals["_PYARROWSCHEMA_TABLESCHEMA_SCHEMAENTRY"]._serialized_end = 1512
    _globals["_PYARROWSCHEMA_COLUMNSCHEMA"]._serialized_start = 1514
    _globals["_PYARROWSCHEMA_COLUMNSCHEMA"]._serialized_end = 1625
    _globals["_PLAN"]._serialized_start = 1628
    _globals["_PLAN"]._serialized_end = 1786
    _globals["_NODE"]._serialized_start = 1788
    _globals["_NODE"]._serialized_end = 1884
    _globals["_NODEIMPL"]._serialized_start = 1887
    _globals["_NODEIMPL"]._serialized_end = 2230
    _globals["_UNKNOWNNODE"]._serialized_start = 2232
    _globals["_UNKNOWNNODE"]._serialized_end = 2274
    _globals["_GIVENSSCAN"]._serialized_start = 2276
    _globals["_GIVENSSCAN"]._serialized_end = 2334
    _globals["_CHALKPROJECT"]._serialized_start = 2336
    _globals["_CHALKPROJECT"]._serialized_end = 2456
    _globals["_PROJECT"]._serialized_start = 2458
    _globals["_PROJECT"]._serialized_end = 2513
    _globals["_DEFAULTINJECTOR"]._serialized_start = 2515
    _globals["_DEFAULTINJECTOR"]._serialized_end = 2603
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/online_query.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.common.v1 import chalk_error_pb2 as chalk_dot_common_dot_v1_dot_chalk__error__pb2
from chalk._gen.chalk.expression.v1 import expression_pb2 as chalk_dot_expression_dot_v1_dot_expression__pb2
from chalk._gen.chalk.graph.v1 import graph_pb2 as chalk_dot_graph_dot_v1_dot_graph__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n"chalk/common/v1/online_query.proto\x12\x0f\x63halk.common.v1\x1a!chalk/common/v1/chalk_error.proto\x1a$chalk/expression/v1/expression.proto\x1a\x1a\x63halk/graph/v1/graph.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xc9\x04\n\x12OnlineQueryRequest\x12G\n\x06inputs\x18\x01 \x03(\x0b\x32/.chalk.common.v1.OnlineQueryRequest.InputsEntryR\x06inputs\x12\x35\n\x07outputs\x18\x02 \x03(\x0b\x32\x1b.chalk.common.v1.OutputExprR\x07outputs\x12\x31\n\x03now\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R\x03now\x88\x01\x01\x12P\n\tstaleness\x18\x04 \x03(\x0b\x32\x32.chalk.common.v1.OnlineQueryRequest.StalenessEntryR\tstaleness\x12=\n\x07\x63ontext\x18\x05 \x01(\x0b\x32#.chalk.common.v1.OnlineQueryContextR\x07\x63ontext\x12V\n\x10response_options\x18\x06 \x01(\x0b\x32+.chalk.common.v1.OnlineQueryResponseOptionsR\x0fresponseOptions\x1aQ\n\x0bInputsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x1a<\n\x0eStalenessEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x06\n\x04_now"\x8e\x04\n\x16OnlineQueryBulkRequest\x12%\n\x0einputs_feather\x18\x01 \x01(\x0cR\rinputsFeather\x12\x35\n\x07outputs\x18\x02 \x03(\x0b\x32\x1b.chalk.common.v1.OutputExprR\x07outputs\x12,\n\x03now\x18\x03 \x03(\x0b\x32\x1a.google.protobuf.TimestampR\x03now\x12T\n\tstaleness\x18\x04 \x03(\x0b\x32\x36.chalk.common.v1.OnlineQueryBulkRequest.StalenessEntryR\tstaleness\x12=\n\x07\x63ontext\x18\x05 \x01(\x0b\x32#.chalk.common.v1.OnlineQueryContextR\x07\x63ontext\x12V\n\x10response_options\x18\x06 \x01(\x0b\x32+.chalk.common.v1.OnlineQueryResponseOptionsR\x0fresponseOptions\x12=\n\tbody_type\x18\x07 \x01(\x0e\x32 .chalk.common.v1.FeatherBodyTypeR\x08\x62odyType\x1a<\n\x0eStalenessEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01"\xb9\x01\n\x12GenericSingleQuery\x12L\n\x0esingle_request\x18\x01 \x01(\x0b\x32#.chalk.common.v1.OnlineQueryRequestH\x00R\rsingleRequest\x12L\n\x0c\x62ulk_request\x18\x02 \x01(\x0b\x32\'.chalk.common.v1.OnlineQueryBulkRequestH\x00R\x0b\x62ulkRequestB\x07\n\x05query"X\n\x17OnlineQueryMultiRequest\x12=\n\x07queries\x18\x01 \x03(\x0b\x32#.chalk.common.v1.GenericSingleQueryR\x07queries"\x99\x01\n\x11\x46\x65\x61tureExpression\x12,\n\x12output_column_name\x18\x01 \x01(\tR\x10outputColumnName\x12\x1c\n\tnamespace\x18\x03 \x01(\tR\tnamespace\x12\x38\n\x04\x65xpr\x18\x02 \x01(\x0b\x32$.chalk.expression.v1.LogicalExprNodeR\x04\x65xpr"\x8c\x01\n\nOutputExpr\x12!\n\x0b\x66\x65\x61ture_fqn\x18\x01 \x01(\tH\x00R\nfeatureFqn\x12S\n\x12\x66\x65\x61ture_expression\x18\x02 \x01(\x0b\x32".chalk.common.v1.FeatureExpressionH\x00R\x11\x66\x65\x61tureExpressionB\x06\n\x04\x65xpr"\xc2\x07\n\x12OnlineQueryContext\x12$\n\x0b\x65nvironment\x18\x01 \x01(\tB\x02\x18\x01R\x0b\x65nvironment\x12\x12\n\x04tags\x18\x02 \x03(\tR\x04tags\x12\x34\n\x16required_resolver_tags\x18\x03 \x03(\tR\x14requiredResolverTags\x12,\n\rdeployment_id\x18\x04 \x01(\tB\x02\x18\x01H\x00R\x0c\x64\x65ploymentId\x88\x01\x01\x12$\n\tbranch_id\x18\x05 \x01(\tB\x02\x18\x01H\x01R\x08\x62ranchId\x88\x01\x01\x12*\n\x0e\x63orrelation_id\x18\x06 \x01(\tH\x02R\rcorrelationId\x88\x01\x01\x12"\n\nquery_name\x18\x07 \x01(\tH\x03R\tqueryName\x88\x01\x01\x12\x31\n\x12query_name_version\x18\x08 \x01(\tH\x04R\x10queryNameVersion\x88\x01\x01\x12J\n\x07options\x18\t \x03(\x0b\x32\x30.chalk.common.v1.OnlineQueryContext.OptionsEntryR\x07options\x12]\n\x1dvalue_metrics_tag_by_features\x18\n \x03(\x0b\x32\x1b.chalk.common.v1.OutputExprR\x19valueMetricsTagByFeatures\x12Z\n\rquery_context\x18\x0b \x03(\x0b\x32\x35.chalk.common.v1.OnlineQueryContext.QueryContextEntryR\x0cqueryContext\x12\x46\n\roverlay_graph\x18\x0c \x01(\x0b\x32\x1c.chalk.graph.v1.OverlayGraphH\x05R\x0coverlayGraph\x88\x01\x01\x1aR\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x1aW\n\x11QueryContextEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x42\x10\n\x0e_deployment_idB\x0c\n\n_branch_idB\x11\n\x0f_correlation_idB\r\n\x0b_query_nameB\x15\n\x13_query_name_versionB\x10\n\x0e_overlay_graph"\xe2\x02\n\x1aOnlineQueryResponseOptions\x12!\n\x0cinclude_meta\x18\x01 \x01(\x08R\x0bincludeMeta\x12\x39\n\x07\x65xplain\x18\x02 \x01(\x0b\x32\x1f.chalk.common.v1.ExplainOptionsR\x07\x65xplain\x12R\n\x10\x65ncoding_options\x18\x03 \x01(\x0b\x32\'.chalk.common.v1.FeatureEncodingOptionsR\x0f\x65ncodingOptions\x12U\n\x08metadata\x18\x04 \x03(\x0b\x32\x39.chalk.common.v1.OnlineQueryResponseOptions.MetadataEntryR\x08metadata\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01"\x10\n\x0e\x45xplainOptions"S\n\x16\x46\x65\x61tureEncodingOptions\x12\x39\n\x19\x65ncode_structs_as_objects\x18\x01 \x01(\x08R\x16\x65ncodeStructsAsObjects"\xcd\x01\n\x13OnlineQueryResponse\x12\x36\n\x04\x64\x61ta\x18\x01 \x01(\x0b\x32".chalk.common.v1.OnlineQueryResultR\x04\x64\x61ta\x12\x33\n\x06\x65rrors\x18\x02 \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorR\x06\x65rrors\x12I\n\rresponse_meta\x18\x03 \x01(\x0b\x32$.chalk.common.v1.OnlineQueryMetadataR\x0cresponseMeta"\xd6\x02\n\x17OnlineQueryBulkResponse\x12!\n\x0cscalars_data\x18\x01 \x01(\x0cR\x0bscalarsData\x12Y\n\x0bgroups_data\x18\x02 \x03(\x0b\x32\x38.chalk.common.v1.OnlineQueryBulkResponse.GroupsDataEntryR\ngroupsData\x12\x33\n\x06\x65rrors\x18\x03 \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorR\x06\x65rrors\x12I\n\rresponse_meta\x18\x04 \x01(\x0b\x32$.chalk.common.v1.OnlineQueryMetadataR\x0cresponseMeta\x1a=\n\x0fGroupsDataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value:\x02\x38\x01"\xc2\x01\n\x15GenericSingleResponse\x12O\n\x0fsingle_response\x18\x01 \x01(\x0b\x32$.chalk.common.v1.OnlineQueryResponseH\x00R\x0esingleResponse\x12O\n\rbulk_response\x18\x02 \x01(\x0b\x32(.chalk.common.v1.OnlineQueryBulkResponseH\x00R\x0c\x62ulkResponseB\x07\n\x05query"\x95\x01\n\x18OnlineQueryMultiResponse\x12\x44\n\tresponses\x18\x01 \x03(\x0b\x32&.chalk.common.v1.GenericSingleResponseR\tresponses\x12\x33\n\x06\x65rrors\x18\x02 \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorR\x06\x65rrors"M\n\x11OnlineQueryResult\x12\x38\n\x07results\x18\x01 \x03(\x0b\x32\x1e.chalk.common.v1.FeatureResultR\x07results"\xb9\x02\n\rFeatureResult\x12\x14\n\x05\x66ield\x18\x01 \x01(\tR\x05\x66ield\x12*\n\x04pkey\x18\x06 \x01(\x0b\x32\x16.google.protobuf.ValueR\x04pkey\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value\x12\x36\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x1b.chalk.common.v1.ChalkErrorH\x00R\x05\x65rror\x88\x01\x01\x12/\n\x02ts\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x01R\x02ts\x88\x01\x01\x12\x35\n\x04meta\x18\x05 \x01(\x0b\x32\x1c.chalk.common.v1.FeatureMetaH\x02R\x04meta\x88\x01\x01\x42\x08\n\x06_errorB\x05\n\x03_tsB\x07\n\x05_meta"\x9b\x01\n\x0b\x46\x65\x61tureMeta\x12.\n\x13\x63hosen_resolver_fqn\x18\x01 \x01(\tR\x11\x63hosenResolverFqn\x12\x1b\n\tcache_hit\x18\x02 \x01(\x08R\x08\x63\x61\x63heHit\x12%\n\x0eprimitive_type\x18\x03 \x01(\tR\rprimitiveType\x12\x18\n\x07version\x18\x04 \x01(\x03R\x07version"\xfa\x05\n\x13OnlineQueryMetadata\x12H\n\x12\x65xecution_duration\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11\x65xecutionDuration\x12#\n\rdeployment_id\x18\x02 \x01(\tR\x0c\x64\x65ploymentId\x12%\n\x0e\x65nvironment_id\x18\x03 \x01(\tR\renvironmentId\x12)\n\x10\x65nvironment_name\x18\x04 \x01(\tR\x0f\x65nvironmentName\x12\x19\n\x08query_id\x18\x05 \x01(\tR\x07queryId\x12\x43\n\x0fquery_timestamp\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0equeryTimestamp\x12\x1d\n\nquery_hash\x18\x07 \x01(\tR\tqueryHash\x12H\n\x0e\x65xplain_output\x18\x08 \x01(\x0b\x32!.chalk.common.v1.QueryExplainInfoR\rexplainOutput\x12N\n\x08metadata\x18\t \x03(\x0b\x32\x32.chalk.common.v1.OnlineQueryMetadata.MetadataEntryR\x08metadata\x12m\n\x13\x61\x64\x64itional_metadata\x18\n \x03(\x0b\x32<.chalk.common.v1.OnlineQueryMetadata.AdditionalMetadataEntryR\x12\x61\x64\x64itionalMetadata\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a]\n\x17\x41\x64\x64itionalMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01"H\n\x10QueryExplainInfo\x12$\n\x0bplan_string\x18\x01 \x01(\tH\x00R\nplanString\x88\x01\x01\x42\x0e\n\x0c_plan_string"\x81\x01\n\x19UploadFeaturesBulkRequest\x12%\n\x0einputs_feather\x18\x01 \x01(\x0cR\rinputsFeather\x12=\n\tbody_type\x18\x07 \x01(\x0e\x32 .chalk.common.v1.FeatherBodyTypeR\x08\x62odyType"Q\n\x1aUploadFeaturesBulkResponse\x12\x33\n\x06\x65rrors\x18\x01 \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorR\x06\x65rrors*w\n\x0f\x46\x65\x61therBodyType\x12!\n\x1d\x46\x45\x41THER_BODY_TYPE_UNSPECIFIED\x10\x00\x12\x1b\n\x17\x46\x45\x41THER_BODY_TYPE_TABLE\x10\x01\x12$\n FEATHER_BODY_TYPE_RECORD_BATCHES\x10\x02\x42\x85\x01\n\x13\x63om.chalk.common.v1B\x10OnlineQueryProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.online_query_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\020OnlineQueryProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_ONLINEQUERYREQUEST_INPUTSENTRY"]._options = None
    _globals["_ONLINEQUERYREQUEST_INPUTSENTRY"]._serialized_options = b"8\001"
    _globals["_ONLINEQUERYREQUEST_STALENESSENTRY"]._options = None
    _globals["_ONLINEQUERYREQUEST_STALENESSENTRY"]._serialized_options = b"8\001"
    _globals["_ONLINEQUERYBULKREQUEST_STALENESSENTRY"]._options = None
    _globals["_ONLINEQUERYBULKREQUEST_STALENESSENTRY"]._serialized_options = b"8\001"
    _globals["_ONLINEQUERYCONTEXT_OPTIONSENTRY"]._options = None
    _globals["_ONLINEQUERYCONTEXT_OPTIONSENTRY"]._serialized_options = b"8\001"
    _globals["_ONLINEQUERYCONTEXT_QUERYCONTEXTENTRY"]._options = None
    _globals["_ONLINEQUERYCONTEXT_QUERYCONTEXTENTRY"]._serialized_options = b"8\001"
    _globals["_ONLINEQUERYCONTEXT"].fields_by_name["environment"]._options = None
    _globals["_ONLINEQUERYCONTEXT"].fields_by_name["environment"]._serialized_options = b"\030\001"
    _globals["_ONLINEQUERYCONTEXT"].fields_by_name["deployment_id"]._options = None
    _globals["_ONLINEQUERYCONTEXT"].fields_by_name["deployment_id"]._serialized_options = b"\030\001"
    _globals["_ONLINEQUERYCONTEXT"].fields_by_name["branch_id"]._options = None
    _globals["_ONLINEQUERYCONTEXT"].fields_by_name["branch_id"]._serialized_options = b"\030\001"
    _globals["_ONLINEQUERYRESPONSEOPTIONS_METADATAENTRY"]._options = None
    _globals["_ONLINEQUERYRESPONSEOPTIONS_METADATAENTRY"]._serialized_options = b"8\001"
    _globals["_ONLINEQUERYBULKRESPONSE_GROUPSDATAENTRY"]._options = None
    _globals["_ONLINEQUERYBULKRESPONSE_GROUPSDATAENTRY"]._serialized_options = b"8\001"
    _globals["_ONLINEQUERYMETADATA_METADATAENTRY"]._options = None
    _globals["_ONLINEQUERYMETADATA_METADATAENTRY"]._serialized_options = b"8\001"
    _globals["_ONLINEQUERYMETADATA_ADDITIONALMETADATAENTRY"]._options = None
    _globals["_ONLINEQUERYMETADATA_ADDITIONALMETADATAENTRY"]._serialized_options = b"8\001"
    _globals["_FEATHERBODYTYPE"]._serialized_start = 5879
    _globals["_FEATHERBODYTYPE"]._serialized_end = 5998
    _globals["_ONLINEQUERYREQUEST"]._serialized_start = 252
    _globals["_ONLINEQUERYREQUEST"]._serialized_end = 837
    _globals["_ONLINEQUERYREQUEST_INPUTSENTRY"]._serialized_start = 686
    _globals["_ONLINEQUERYREQUEST_INPUTSENTRY"]._serialized_end = 767
    _globals["_ONLINEQUERYREQUEST_STALENESSENTRY"]._serialized_start = 769
    _globals["_ONLINEQUERYREQUEST_STALENESSENTRY"]._serialized_end = 829
    _globals["_ONLINEQUERYBULKREQUEST"]._serialized_start = 840
    _globals["_ONLINEQUERYBULKREQUEST"]._serialized_end = 1366
    _globals["_ONLINEQUERYBULKREQUEST_STALENESSENTRY"]._serialized_start = 769
    _globals["_ONLINEQUERYBULKREQUEST_STALENESSENTRY"]._serialized_end = 829
    _globals["_GENERICSINGLEQUERY"]._serialized_start = 1369
    _globals["_GENERICSINGLEQUERY"]._serialized_end = 1554
    _globals["_ONLINEQUERYMULTIREQUEST"]._serialized_start = 1556
    _globals["_ONLINEQUERYMULTIREQUEST"]._serialized_end = 1644
    _globals["_FEATUREEXPRESSION"]._serialized_start = 1647
    _globals["_FEATUREEXPRESSION"]._serialized_end = 1800
    _globals["_OUTPUTEXPR"]._serialized_start = 1803
    _globals["_OUTPUTEXPR"]._serialized_end = 1943
    _globals["_ONLINEQUERYCONTEXT"]._serialized_start = 1946
    _globals["_ONLINEQUERYCONTEXT"]._serialized_end = 2908
    _globals["_ONLINEQUERYCONTEXT_OPTIONSENTRY"]._serialized_start = 2630
    _globals["_ONLINEQUERYCONTEXT_OPTIONSENTRY"]._serialized_end = 2712
    _globals["_ONLINEQUERYCONTEXT_QUERYCONTEXTENTRY"]._serialized_start = 2714
    _globals["_ONLINEQUERYCONTEXT_QUERYCONTEXTENTRY"]._serialized_end = 2801
    _globals["_ONLINEQUERYRESPONSEOPTIONS"]._serialized_start = 2911
    _globals["_ONLINEQUERYRESPONSEOPTIONS"]._serialized_end = 3265
    _globals["_ONLINEQUERYRESPONSEOPTIONS_METADATAENTRY"]._serialized_start = 3206
    _globals["_ONLINEQUERYRESPONSEOPTIONS_METADATAENTRY"]._serialized_end = 3265
    _globals["_EXPLAINOPTIONS"]._serialized_start = 3267
    _globals["_EXPLAINOPTIONS"]._serialized_end = 3283
    _globals["_FEATUREENCODINGOPTIONS"]._serialized_start = 3285
    _globals["_FEATUREENCODINGOPTIONS"]._serialized_end = 3368
    _globals["_ONLINEQUERYRESPONSE"]._serialized_start = 3371
    _globals["_ONLINEQUERYRESPONSE"]._serialized_end = 3576
    _globals["_ONLINEQUERYBULKRESPONSE"]._serialized_start = 3579
    _globals["_ONLINEQUERYBULKRESPONSE"]._serialized_end = 3921
    _globals["_ONLINEQUERYBULKRESPONSE_GROUPSDATAENTRY"]._serialized_start = 3860
    _globals["_ONLINEQUERYBULKRESPONSE_GROUPSDATAENTRY"]._serialized_end = 3921
    _globals["_GENERICSINGLERESPONSE"]._serialized_start = 3924
    _globals["_GENERICSINGLERESPONSE"]._serialized_end = 4118
    _globals["_ONLINEQUERYMULTIRESPONSE"]._serialized_start = 4121
    _globals["_ONLINEQUERYMULTIRESPONSE"]._serialized_end = 4270
    _globals["_ONLINEQUERYRESULT"]._serialized_start = 4272
    _globals["_ONLINEQUERYRESULT"]._serialized_end = 4349
    _globals["_FEATURERESULT"]._serialized_start = 4352
    _globals["_FEATURERESULT"]._serialized_end = 4665
    _globals["_FEATUREMETA"]._serialized_start = 4668
    _globals["_FEATUREMETA"]._serialized_end = 4823
    _globals["_ONLINEQUERYMETADATA"]._serialized_start = 4826
    _globals["_ONLINEQUERYMETADATA"]._serialized_end = 5588
    _globals["_ONLINEQUERYMETADATA_METADATAENTRY"]._serialized_start = 3206
    _globals["_ONLINEQUERYMETADATA_METADATAENTRY"]._serialized_end = 3265
    _globals["_ONLINEQUERYMETADATA_ADDITIONALMETADATAENTRY"]._serialized_start = 5495
    _globals["_ONLINEQUERYMETADATA_ADDITIONALMETADATAENTRY"]._serialized_end = 5588
    _globals["_QUERYEXPLAININFO"]._serialized_start = 5590
    _globals["_QUERYEXPLAININFO"]._serialized_end = 5662
    _globals["_UPLOADFEATURESBULKREQUEST"]._serialized_start = 5665
    _globals["_UPLOADFEATURESBULKREQUEST"]._serialized_end = 5794
    _globals["_UPLOADFEATURESBULKRESPONSE"]._serialized_start = 5796
    _globals["_UPLOADFEATURESBULKRESPONSE"]._serialized_end = 5877
# @@protoc_insertion_point(module_scope)



import argparse
import json
import logging as pylogging
import os
import sys
import traceback
from pathlib import Path
from typing import List, Optional, cast

from dataclasses_json import DataClassJsonMixin

from chalk._gen.chalk.artifacts.v1 import export_pb2 as export_pb
from chalk._lsp.error_builder import LSPErrorBuilder
from chalk.config.project_config import load_project_config
from chalk.importer import import_all_files
from chalk.parsed._proto.export import import_files_then_export_from_registry
from chalk.parsed._proto.lsp import convert_lsp_gql_to_proto
from chalk.parsed.user_types_to_json import get_lsp_gql, get_registered_types_as_json
from chalk.utils.stubgen import configure_stubgen_argparse, run_stubgen


def get_list_results(directory: Optional[str], file_allowlist: Optional[List[str]]):
    if directory is not None:
        os.chdir(directory)
    scope_to = Path(directory or os.getcwd())
    try:
        failed = import_all_files(
            file_allowlist=file_allowlist,
            project_root=None if directory is None else Path(directory),
        )
        return get_registered_types_as_json(scope_to, failed)
    except Exception:
        ex_type, ex_value, ex_traceback = sys.exc_info()
        assert ex_type is not None
        relevant_traceback = f"""{os.linesep.join(traceback.format_tb(ex_traceback))}
\n{ex_type.__name__}: {str(ex_value)}
"""
        return json.dumps(
            dict(
                failed=[dict(traceback=relevant_traceback)],
                lsp=cast(DataClassJsonMixin, get_lsp_gql()).to_dict(),
            ),
            indent=2,
        )


def get_artifacts_export_safe(directory: Optional[str], file_allowlist: Optional[List[str]]) -> bytes:
    try:
        return import_files_then_export_from_registry(directory, file_allowlist).SerializeToString()
    except Exception:
        ex_type, ex_value, ex_traceback = sys.exc_info()
        assert ex_type is not None
        relevant_traceback = f"""{os.linesep.join(traceback.format_tb(ex_traceback))}
\n{ex_type.__name__}: {str(ex_value)}
"""
        return export_pb.Export(
            failed=[
                export_pb.FailedImport(
                    traceback=relevant_traceback,
                )
            ],
            lsp=convert_lsp_gql_to_proto(get_lsp_gql()),
        ).SerializeToString()


def dump_cmd(filename: str, directory: Optional[str], filter_file: Optional[str], use_proto: bool = False):
    file_allowlist = None
    if filter_file is not None and os.path.exists(filter_file):
        with open(filter_file) as f:
            file_allowlist = [f.strip() for f in f.readlines()]

    if use_proto:
        with open(filename, "wb") as f:
            f.write(
                get_artifacts_export_safe(
                    directory=directory,
                    file_allowlist=file_allowlist,
                )
            )
    else:
        with open(filename, "w") as f:
            f.write(
                get_list_results(
                    directory=directory,
                    file_allowlist=file_allowlist,
                )
            )


def config_cmd():
    try:
        model = load_project_config()
    except ValueError as e:
        print(json.dumps({"error": str(e)}))
        return

    if model is None:
        print("No `chalk.yaml` configuration file found")
        return
    json_response = model.as_dict()
    print(json.dumps(json_response))


def cli(args_override: Optional[List[str]] = None):
    parser = argparse.ArgumentParser(
        prog="Chalk Python CLI",
        description=(
            "You typically do not need to invoke this utility directly. "
            "Prefer https://github.com/chalk-ai/cli instead."
        ),
    )
    parser.add_argument("--log-level", help="Print debug info", nargs="?")
    subparsers = parser.add_subparsers(dest="command")

    # CONFIG
    subparsers.add_parser("config", help="Print the config for the current project")

    # EXPORT
    export_parser = subparsers.add_parser(
        "export", help="Write the resolvers for the current project to the given file"
    )
    export_parser.add_argument("filename", help="Write to this file")
    export_parser.add_argument("--directory", help="Scope to this directory", nargs="?")
    export_parser.add_argument("--file_filter", help="Path containing only files to consider", nargs="?")
    # >= 3.9 version
    # export_parser.add_argument(
    #     "--lsp",
    #     help="If set, will output the diagnostics in the format expected by the LSP instead of eagerly raising",
    #     action=argparse.BooleanOptionalAction,
    # )
    # < 3.9 version
    export_parser.add_argument("--lsp", action="store_true")
    export_parser.add_argument("--no-lsp", dest="lsp", action="store_false")
    export_parser.set_defaults(lsp=False)

    export_parser.add_argument("--proto", action="store_true")
    export_parser.set_defaults(proto=False)

    # STUBGEN
    stubgen_parser = subparsers.add_parser("stubgen", help="Generate type stubs for feature set classes")
    configure_stubgen_argparse(stubgen_parser)

    # Parsing only known args for forwards compatibility.
    # Changing this to `.parse_args` means once the args are
    # set for a command, you can never add to them.
    # Please do not change to `.parse_args`.
    args, _ = parser.parse_known_args(args_override)
    if args.log_level:
        level = getattr(pylogging, args.log_level.upper())
        pylogging.basicConfig(
            format="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%m/%d/%Y %I:%M:%S %p",
            level=level,
        )

    if args.command == "export":
        LSPErrorBuilder.lsp = args.lsp
        dump_cmd(filename=args.filename, directory=args.directory, filter_file=args.file_filter, use_proto=args.proto)

    elif args.command == "config":
        config_cmd()

    elif args.command == "stubgen":
        run_stubgen(args=args, file_filter=args.file_filter)

    else:
        parser.print_help(sys.stderr)
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(cli())



from chalk._gen.chalk.aggregate.v1 import service_pb2 as _service_pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.server.v1 import datasets_pb2 as _datasets_pb2
from chalk._gen.chalk.server.v1 import performance_summary_pb2 as _performance_summary_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class OfflineQueryStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    OFFLINE_QUERY_STATUS_UNSPECIFIED: _ClassVar[OfflineQueryStatus]
    OFFLINE_QUERY_STATUS_UNKNOWN: _ClassVar[OfflineQueryStatus]
    OFFLINE_QUERY_STATUS_WORKING: _ClassVar[OfflineQueryStatus]
    OFFLINE_QUERY_STATUS_FAILED: _ClassVar[OfflineQueryStatus]
    OFFLINE_QUERY_STATUS_COMPLETED: _ClassVar[OfflineQueryStatus]
    OFFLINE_QUERY_STATUS_CANCELED: _ClassVar[OfflineQueryStatus]
    OFFLINE_QUERY_STATUS_QUEUED: _ClassVar[OfflineQueryStatus]

class OfflineQueryKind(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    OFFLINE_QUERY_KIND_UNSPECIFIED: _ClassVar[OfflineQueryKind]
    OFFLINE_QUERY_KIND_UNKNOWN: _ClassVar[OfflineQueryKind]
    OFFLINE_QUERY_KIND_ASYNC_OFFLINE_QUERY: _ClassVar[OfflineQueryKind]
    OFFLINE_QUERY_KIND_CRON_OFFLINE_QUERY: _ClassVar[OfflineQueryKind]
    OFFLINE_QUERY_KIND_OFFLINE_QUERY: _ClassVar[OfflineQueryKind]
    OFFLINE_QUERY_KIND_DATASET_INGESTION: _ClassVar[OfflineQueryKind]
    OFFLINE_QUERY_KIND_AGGREGATION_BACKFILL: _ClassVar[OfflineQueryKind]

OFFLINE_QUERY_STATUS_UNSPECIFIED: OfflineQueryStatus
OFFLINE_QUERY_STATUS_UNKNOWN: OfflineQueryStatus
OFFLINE_QUERY_STATUS_WORKING: OfflineQueryStatus
OFFLINE_QUERY_STATUS_FAILED: OfflineQueryStatus
OFFLINE_QUERY_STATUS_COMPLETED: OfflineQueryStatus
OFFLINE_QUERY_STATUS_CANCELED: OfflineQueryStatus
OFFLINE_QUERY_STATUS_QUEUED: OfflineQueryStatus
OFFLINE_QUERY_KIND_UNSPECIFIED: OfflineQueryKind
OFFLINE_QUERY_KIND_UNKNOWN: OfflineQueryKind
OFFLINE_QUERY_KIND_ASYNC_OFFLINE_QUERY: OfflineQueryKind
OFFLINE_QUERY_KIND_CRON_OFFLINE_QUERY: OfflineQueryKind
OFFLINE_QUERY_KIND_OFFLINE_QUERY: OfflineQueryKind
OFFLINE_QUERY_KIND_DATASET_INGESTION: OfflineQueryKind
OFFLINE_QUERY_KIND_AGGREGATION_BACKFILL: OfflineQueryKind

class OfflineQueryMeta(_message.Message):
    __slots__ = (
        "id",
        "operation_id",
        "environment_id",
        "deployment_id",
        "created_at",
        "query_meta",
        "query_plan_id",
        "branch_name",
        "dataset_id",
        "dataset_name",
        "has_errors",
        "agent_id",
        "trace_id",
        "correlation_id",
        "completed_at",
        "status",
        "has_plan_stages",
        "total_computers",
        "num_completed_computers",
        "total_partitions",
        "num_completed_partitions",
        "recompute_features",
        "spine_sql_query",
        "filters",
        "planner_options",
        "invoker_options",
        "query_type",
        "tags",
        "required_resolver_tags",
        "aggregate_backfill_id",
        "output",
        "required_output",
        "raw_body_filename",
        "dataset_revision",
        "time_series",
        "evaluation_run_id",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    QUERY_META_FIELD_NUMBER: _ClassVar[int]
    QUERY_PLAN_ID_FIELD_NUMBER: _ClassVar[int]
    BRANCH_NAME_FIELD_NUMBER: _ClassVar[int]
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    DATASET_NAME_FIELD_NUMBER: _ClassVar[int]
    HAS_ERRORS_FIELD_NUMBER: _ClassVar[int]
    AGENT_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    CORRELATION_ID_FIELD_NUMBER: _ClassVar[int]
    COMPLETED_AT_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    HAS_PLAN_STAGES_FIELD_NUMBER: _ClassVar[int]
    TOTAL_COMPUTERS_FIELD_NUMBER: _ClassVar[int]
    NUM_COMPLETED_COMPUTERS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_PARTITIONS_FIELD_NUMBER: _ClassVar[int]
    NUM_COMPLETED_PARTITIONS_FIELD_NUMBER: _ClassVar[int]
    RECOMPUTE_FEATURES_FIELD_NUMBER: _ClassVar[int]
    SPINE_SQL_QUERY_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    PLANNER_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    INVOKER_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    QUERY_TYPE_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_RESOLVER_TAGS_FIELD_NUMBER: _ClassVar[int]
    AGGREGATE_BACKFILL_ID_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    RAW_BODY_FILENAME_FIELD_NUMBER: _ClassVar[int]
    DATASET_REVISION_FIELD_NUMBER: _ClassVar[int]
    TIME_SERIES_FIELD_NUMBER: _ClassVar[int]
    EVALUATION_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    id: int
    operation_id: str
    environment_id: str
    deployment_id: str
    created_at: _timestamp_pb2.Timestamp
    query_meta: _struct_pb2.Value
    query_plan_id: str
    branch_name: str
    dataset_id: str
    dataset_name: str
    has_errors: bool
    agent_id: str
    trace_id: str
    correlation_id: str
    completed_at: _timestamp_pb2.Timestamp
    status: OfflineQueryStatus
    has_plan_stages: bool
    total_computers: int
    num_completed_computers: int
    total_partitions: int
    num_completed_partitions: int
    recompute_features: str
    spine_sql_query: str
    filters: _struct_pb2.Value
    planner_options: _struct_pb2.Value
    invoker_options: _struct_pb2.Value
    query_type: OfflineQueryKind
    tags: _containers.RepeatedScalarFieldContainer[str]
    required_resolver_tags: _containers.RepeatedScalarFieldContainer[str]
    aggregate_backfill_id: str
    output: _struct_pb2.Value
    required_output: _struct_pb2.Value
    raw_body_filename: str
    dataset_revision: _datasets_pb2.DatasetRevisionMeta
    time_series: _containers.RepeatedCompositeFieldContainer[_service_pb2.PlanAggregateBackfillResponse]
    evaluation_run_id: str
    def __init__(
        self,
        id: _Optional[int] = ...,
        operation_id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        query_meta: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...,
        query_plan_id: _Optional[str] = ...,
        branch_name: _Optional[str] = ...,
        dataset_id: _Optional[str] = ...,
        dataset_name: _Optional[str] = ...,
        has_errors: bool = ...,
        agent_id: _Optional[str] = ...,
        trace_id: _Optional[str] = ...,
        correlation_id: _Optional[str] = ...,
        completed_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        status: _Optional[_Union[OfflineQueryStatus, str]] = ...,
        has_plan_stages: bool = ...,
        total_computers: _Optional[int] = ...,
        num_completed_computers: _Optional[int] = ...,
        total_partitions: _Optional[int] = ...,
        num_completed_partitions: _Optional[int] = ...,
        recompute_features: _Optional[str] = ...,
        spine_sql_query: _Optional[str] = ...,
        filters: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...,
        planner_options: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...,
        invoker_options: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...,
        query_type: _Optional[_Union[OfflineQueryKind, str]] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        required_resolver_tags: _Optional[_Iterable[str]] = ...,
        aggregate_backfill_id: _Optional[str] = ...,
        output: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...,
        required_output: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...,
        raw_body_filename: _Optional[str] = ...,
        dataset_revision: _Optional[_Union[_datasets_pb2.DatasetRevisionMeta, _Mapping]] = ...,
        time_series: _Optional[_Iterable[_Union[_service_pb2.PlanAggregateBackfillResponse, _Mapping]]] = ...,
        evaluation_run_id: _Optional[str] = ...,
    ) -> None: ...

class ListOfflineQueriesRequest(_message.Message):
    __slots__ = (
        "cursor",
        "limit",
        "start_date",
        "end_date",
        "id_filter",
        "agent_id_filter",
        "branch_filter",
        "kind_filter",
        "status_filter",
        "aggregation_backfill_id_filter",
        "evaluation_run_id_filter",
    )
    CURSOR_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    START_DATE_FIELD_NUMBER: _ClassVar[int]
    END_DATE_FIELD_NUMBER: _ClassVar[int]
    ID_FILTER_FIELD_NUMBER: _ClassVar[int]
    AGENT_ID_FILTER_FIELD_NUMBER: _ClassVar[int]
    BRANCH_FILTER_FIELD_NUMBER: _ClassVar[int]
    KIND_FILTER_FIELD_NUMBER: _ClassVar[int]
    STATUS_FILTER_FIELD_NUMBER: _ClassVar[int]
    AGGREGATION_BACKFILL_ID_FILTER_FIELD_NUMBER: _ClassVar[int]
    EVALUATION_RUN_ID_FILTER_FIELD_NUMBER: _ClassVar[int]
    cursor: str
    limit: int
    start_date: str
    end_date: str
    id_filter: str
    agent_id_filter: str
    branch_filter: str
    kind_filter: OfflineQueryKind
    status_filter: OfflineQueryStatus
    aggregation_backfill_id_filter: str
    evaluation_run_id_filter: str
    def __init__(
        self,
        cursor: _Optional[str] = ...,
        limit: _Optional[int] = ...,
        start_date: _Optional[str] = ...,
        end_date: _Optional[str] = ...,
        id_filter: _Optional[str] = ...,
        agent_id_filter: _Optional[str] = ...,
        branch_filter: _Optional[str] = ...,
        kind_filter: _Optional[_Union[OfflineQueryKind, str]] = ...,
        status_filter: _Optional[_Union[OfflineQueryStatus, str]] = ...,
        aggregation_backfill_id_filter: _Optional[str] = ...,
        evaluation_run_id_filter: _Optional[str] = ...,
    ) -> None: ...

class ListOfflineQueriesResponse(_message.Message):
    __slots__ = ("offline_queries", "cursor")
    OFFLINE_QUERIES_FIELD_NUMBER: _ClassVar[int]
    CURSOR_FIELD_NUMBER: _ClassVar[int]
    offline_queries: _containers.RepeatedCompositeFieldContainer[OfflineQueryMeta]
    cursor: str
    def __init__(
        self,
        offline_queries: _Optional[_Iterable[_Union[OfflineQueryMeta, _Mapping]]] = ...,
        cursor: _Optional[str] = ...,
    ) -> None: ...

class GetOfflineQueryRequest(_message.Message):
    __slots__ = ("offline_query_id",)
    OFFLINE_QUERY_ID_FIELD_NUMBER: _ClassVar[int]
    offline_query_id: str
    def __init__(self, offline_query_id: _Optional[str] = ...) -> None: ...

class GetOfflineQueryResponse(_message.Message):
    __slots__ = ("offline_query",)
    OFFLINE_QUERY_FIELD_NUMBER: _ClassVar[int]
    offline_query: OfflineQueryMeta
    def __init__(self, offline_query: _Optional[_Union[OfflineQueryMeta, _Mapping]] = ...) -> None: ...



from __future__ import annotations

import concurrent.futures
import dataclasses
import queue
import typing
import uuid
from datetime import datetime, timedelta, timezone
from threading import Lock
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Mapping, Optional, Sequence

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset
from pyarrow.fs import S3FileSystem

from chalk.clogging import chalk_logger
from chalk.features import Feature
from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.query_execution_parameters import QueryExecutionParameters
from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.utils.df_utils import pa_array_to_pl_series
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.threading import DEFAULT_IO_EXECUTOR, MultiSemaphore
from chalk.utils.tracing import safe_incr, safe_set_gauge, safe_trace

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection
    from sqlalchemy.engine.url import URL

    try:
        from pyathena.connection import Connection as AthenaConnection
    except ImportError:
        pass

_logger = get_logger(__name__)

_WorkerId = typing.NewType("_WorkerId", int)


@dataclasses.dataclass(frozen=True)
class AthenaResultHandle:
    uri: str
    compressed_size: int

    @property
    def estimated_uncompressed_size(self) -> int:
        # this is a (hopefully) conservative estimate.
        return self.compressed_size * 8

    def to_arrow(self, fs: S3FileSystem) -> pa.Table:
        dataset = pyarrow.dataset.dataset(self.uri, format="parquet", filesystem=fs)
        return dataset.to_table()


_ATHENA_AWS_REGION_NAME = "ATHENA_AWS_REGION"
_ATHENA_AWS_ACCESS_KEY_ID_NAME = "ATHENA_AWS_ACCESS_KEY_ID"
_ATHENA_AWS_ACCESS_KEY_SECRET_NAME = "ATHENA_AWS_ACCESS_KEY_SECRET"
_ATHENA_S3_STAGING_DIR_NAME = "ATHENA_S3_STAGING_DIR"
_ATHENA_ROLE_ARN_NAME = "ATHENA_ROLE_ARN"
_ATHENA_SCHEMA_NAME_NAME = "ATHENA_SCHEMA_NAME"
_ATHENA_CATALOG_NAME_NAME = "ATHENA_CATALOG_NAME"
_ATHENA_WORK_GROUP_NAME = "ATHENA_WORK_GROUP"


class AthenaSourceImpl(BaseSQLSource):
    kind = SQLSourceKind.athena

    def __init__(
        self,
        *,
        name: str | None = None,
        aws_region: str | None = None,
        aws_access_key_id: str | None = None,
        aws_access_key_secret: str | None = None,
        s3_staging_dir: str | None = None,
        schema_name: str | None = None,
        catalog_name: str | None = None,
        work_group: str | None = None,
        role_arn: str | None = None,
        engine_args: Dict[str, Any] | None = None,
        executor: Optional[concurrent.futures.ThreadPoolExecutor] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            import pyathena
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[athena]")
        else:
            del pyathena  # unused here

        self.aws_region = aws_region or load_integration_variable(
            integration_name=name, name=_ATHENA_AWS_REGION_NAME, override=integration_variable_override
        )
        self.aws_access_key_id = aws_access_key_id or load_integration_variable(
            integration_name=name, name=_ATHENA_AWS_ACCESS_KEY_ID_NAME, override=integration_variable_override
        )
        self.aws_access_key_secret = aws_access_key_secret or load_integration_variable(
            integration_name=name, name=_ATHENA_AWS_ACCESS_KEY_SECRET_NAME, override=integration_variable_override
        )
        self.s3_staging_dir = s3_staging_dir or load_integration_variable(
            integration_name=name, name=_ATHENA_S3_STAGING_DIR_NAME, override=integration_variable_override
        )
        self.role_arn = role_arn or load_integration_variable(
            integration_name=name, name=_ATHENA_ROLE_ARN_NAME, override=integration_variable_override
        )
        self.schema_name = schema_name or load_integration_variable(
            integration_name=name, name=_ATHENA_SCHEMA_NAME_NAME, override=integration_variable_override
        )
        self.catalog_name = catalog_name or load_integration_variable(
            integration_name=name, name=_ATHENA_CATALOG_NAME_NAME, override=integration_variable_override
        )
        self.work_group = work_group or load_integration_variable(
            integration_name=name, name=_ATHENA_WORK_GROUP_NAME, override=integration_variable_override
        )
        self.executor = executor or DEFAULT_IO_EXECUTOR

        self._cached_s3_client = None
        self._s3_client_expiration = None

        self._cached_s3_filesystem = None
        self._s3_filesystem_expiration = None
        self._s3fs_lock = Lock()

        if engine_args is None:
            engine_args = {}
        engine_args.setdefault("pool_size", 20)
        engine_args.setdefault("max_overflow", 60)
        engine_args.setdefault("connect_args", {"s3_staging_dir": s3_staging_dir})

        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args={})

    def _pyathena_connection(self) -> AthenaConnection:
        try:
            from pyathena.arrow.async_cursor import AsyncArrowCursor
            from pyathena.connection import Connection as AthenaConnection
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[athena]")
        return AthenaConnection(
            s3_staging_dir=self.s3_staging_dir,
            role_arn=self.role_arn,
            schema_name=self.schema_name or "default",
            catalog_name=self.catalog_name or "awsdatacatalog",
            work_group=self.work_group,
            region_name=self.aws_region,
            access_key=self.aws_access_key_id,
            secret_key=self.aws_access_key_secret,
            cursor_class=AsyncArrowCursor,
        )

    def supports_inefficient_fallback(self) -> bool:
        return False

    def get_sqlglot_dialect(self) -> str | None:
        # Athena was introduced to sqlglot in v22.3.0: it does seem sqlglot version changes have historically
        # broken query pushdown so defer to trino for now
        return "trino"

    def _create_s3_client(self):
        import boto3

        assumed_aws_client_id = None
        assumed_aws_client_secret = None
        assumed_session_token = None
        if self.role_arn:
            _logger.debug(
                f"Attempting to assume arn {self.role_arn=}, (aws region override = {self.aws_region}) with a web identity"
            )
            sts_client = boto3.client("sts")
            response = sts_client.assume_role(RoleArn=self.role_arn, RoleSessionName="chalk-athena-s3-assume")
            if not response or "Credentials" not in response:
                raise ValueError(f"Failed to assume role in sts client: {self.role_arn}")
            assumed_aws_client_id = response["Credentials"]["AccessKeyId"]
            assumed_aws_client_secret = response["Credentials"]["SecretAccessKey"]
            assumed_session_token = response["Credentials"]["SessionToken"]
            if "Expiration" in response["Credentials"]:
                self._s3_client_expiration = response["Credentials"]["Expiration"]

        self._cached_s3_client = boto3.client(
            "s3",
            region_name=self.aws_region,
            aws_access_key_id=self.aws_access_key_id or assumed_aws_client_id,
            aws_secret_access_key=self.aws_access_key_secret or assumed_aws_client_secret,
            aws_session_token=assumed_session_token,
        )

    @property
    def _s3_client(self):
        if (
            self._cached_s3_client is None
            or self._s3_client_expiration is None
            or self._s3_client_expiration < datetime.now(tz=timezone.utc) + timedelta(minutes=5)
        ):
            self._create_s3_client()
        assert self._cached_s3_client is not None
        return self._cached_s3_client

    def _create_s3_filesystem(self):
        import boto3

        assumed_aws_client_id = None
        assumed_aws_client_secret = None
        assumed_session_token = None
        if self.role_arn:
            _logger.debug(
                f"Attempting to assume arn {self.role_arn=}, (aws region override = {self.aws_region}) with a web identity"
            )
            sts_client = boto3.client("sts")
            response = sts_client.assume_role(RoleArn=self.role_arn, RoleSessionName="chalk-athena-s3-assume")
            if not response or "Credentials" not in response:
                raise ValueError(f"Failed to assume role in sts client: {self.role_arn}")
            assumed_aws_client_id = response["Credentials"]["AccessKeyId"]
            assumed_aws_client_secret = response["Credentials"]["SecretAccessKey"]
            assumed_session_token = response["Credentials"]["SessionToken"]
            if "Expiration" in response["Credentials"]:
                self._s3_filesystem_expiration = response["Credentials"]["Expiration"]

        self._cached_s3_filesystem = S3FileSystem(
            region=self.aws_region,
            access_key=self.aws_access_key_id or assumed_aws_client_id,
            secret_key=self.aws_access_key_secret or assumed_aws_client_secret,
            session_token=assumed_session_token,
        )

    def _s3_filesystem(self):
        with self._s3fs_lock:
            if (
                self._cached_s3_filesystem is None
                or self._s3_filesystem_expiration is None
                or self._s3_filesystem_expiration < datetime.now(tz=timezone.utc) + timedelta(minutes=5)
            ):
                self._create_s3_filesystem()
            assert self._cached_s3_filesystem is not None
            return self._cached_s3_filesystem

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        query = {
            k: v
            for k, v in {
                "s3_staging_dir": self.s3_staging_dir,
                "role_arn": self.role_arn,
                "work_group": self.work_group,
                "unload": "true",
            }.items()
            if v is not None
        }
        # https://laughingman7743.github.io/PyAthena/sqlalchemy.html
        return URL.create(
            drivername="awsathena+arrow",
            username=self.aws_access_key_id,
            password=self.aws_access_key_secret,
            host=f"athena.{self.aws_region}.amazonaws.com",
            port=443,
            database=self.schema_name,
            query=query,
        )

    @staticmethod
    def _rewrite_query_for_unload(sql: str, s3_staging_dir_for_job: str):
        rewritten_sql = sql.rstrip(";\n ")
        new_query = f"""
        UNLOAD ({rewritten_sql}) TO '{s3_staging_dir_for_job}' WITH (format = 'PARQUET', compression = 'SNAPPY')
        """
        return new_query

    def _execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[Connection],
        query_execution_parameters: QueryExecutionParameters,
    ) -> Iterable[pa.RecordBatch]:
        with safe_trace("athena.execute_query_efficient"):
            try:
                from pyathena.arrow.async_cursor import AsyncArrowCursor
                from pyathena.arrow.result_set import AthenaArrowResultSet
            except ModuleNotFoundError:
                raise missing_dependency_exception("chalkpy[athena]")

            assert len(finalized_query.temp_tables) == 0, "Should not create temp tables with athena source"
            if self.s3_staging_dir is None:
                raise ValueError("Could not query Athena, no s3_staging_dir set")

            formatted_op, positional_params, named_params = self.compile_query(finalized_query)
            assert (
                len(positional_params) == 0 or len(named_params) == 0
            ), "Should not mix positional and named parameters"
            execution_params = None
            paramstyle = None
            if len(positional_params) > 0:
                execution_params = list(positional_params)
                if not all(isinstance(x, str) for x in positional_params):
                    raise ValueError("Only strings are allowed as positional parameters in Athena client")
            elif len(named_params) > 0:
                execution_params = named_params
                paramstyle = "named"
            result_handles: queue.Queue[
                AthenaResultHandle
            ] = queue.Queue()  # Using a queue since we'll be concurrently reading
            with self._pyathena_connection().cursor() as cursor:
                job_id = str(uuid.uuid4())
                job_prefix = f"chalk-unload/{job_id}"
                s3_prefix = f"{self.s3_staging_dir.rstrip('/')}/{job_prefix}"
                assert isinstance(cursor, AsyncArrowCursor)

                final_sql = self._rewrite_query_for_unload(
                    sql=formatted_op,
                    s3_staging_dir_for_job=s3_prefix,
                )

                query_id, query_fut = cursor.execute(
                    operation=final_sql,
                    parameters=execution_params,
                    paramstyle=paramstyle,
                )

                query_result = query_fut.result()
                assert isinstance(
                    query_result, AthenaArrowResultSet
                ), "Expected athena query result to be AthenaArrowResultSet"
                if query_result.error_type and query_result.error_category and query_result.error_message:
                    chalk_logger.error(
                        f"Failed to execute Athena unload query. Error info: Type: {query_result.error_type}, Category: {query_result.error_category}, Message: {query_result.error_message}"
                    )
                    raise ValueError(
                        f"Failed to execute Athena unload query. Error info: Type: {query_result.error_type}, Category: {query_result.error_category}, Message: {query_result.error_message}"
                    )

                chalk_logger.info(
                    f"Executed Athena unload query successfully. Query ID: {query_id}",
                )
                bucket_name = s3_prefix.split("/")[2]
                remaining_prefix = "/".join(s3_prefix.split("/")[3:])
                objects_list_response = self._s3_client.list_objects_v2(Bucket=bucket_name, Prefix=remaining_prefix)
                if "Contents" not in objects_list_response:
                    chalk_logger.error(
                        f"Failed to enumerate unloaded files for Athena query with query ID: {query_id}. This may mean there was no data to unload."
                    )
                    raise ValueError(
                        "Failed to enumerate unloaded files for Athena query: this can happen if no data was unloaded."
                    )

                chalk_logger.info(f"Found {len(objects_list_response['Contents'])} unloaded files")
                for object in objects_list_response["Contents"]:
                    if "Key" not in object or "Size" not in object:
                        raise ValueError(f"Expected 'Key' and 'Size' in Athena unload response: {object}")
                    object_key = object["Key"]
                    chalk_logger.info(f"Found unloaded file: {object_key}")
                    result_handles.put_nowait(
                        AthenaResultHandle(uri=f"{bucket_name}/{object_key}", compressed_size=object["Size"])
                    )

                yield from self._yield_from_result_handles(
                    result_handles=result_handles,
                    query_execution_parameters=query_execution_parameters,
                    columns_to_features=columns_to_features,
                )

    def _postprocess_table(self, features: Mapping[str, Feature], tbl: pa.Table):
        columns: list[pa.Array] = []
        column_names: list[str] = []
        chalk_logger.info(
            f"Received a PyArrow table from Athena with {len(tbl)} rows; {len(tbl.column_names)} columns; {tbl.nbytes=}; {tbl.schema=}"
        )

        for col_name, feature in features.items():
            try:
                column = tbl[col_name]
                expected_type = feature.converter.pyarrow_dtype
                actual_type = tbl.schema.field(col_name).type
                if pa.types.is_list(expected_type) or pa.types.is_large_list(expected_type):
                    if pa.types.is_string(actual_type) or pa.types.is_large_string(actual_type):
                        series = pa_array_to_pl_series(tbl[col_name])
                        column = series.str.json_extract(feature.converter.polars_dtype).to_arrow().cast(expected_type)
                if actual_type != expected_type:
                    column = column.cast(options=pc.CastOptions(target_type=expected_type, allow_time_truncate=True))
                if isinstance(column, pa.ChunkedArray):
                    column = column.combine_chunks()
                columns.append(column)
                column_names.append(feature.root_fqn)
            except:
                chalk_logger.error(f"Failed to deserialize column '{col_name}' into '{feature}'", exc_info=True)
                raise

        return pa.RecordBatch.from_arrays(arrays=columns, names=column_names)

    def _download_worker(
        self,
        result_handles: queue.Queue[AthenaResultHandle],
        sem: MultiSemaphore | None,
        pa_table_queue: queue.Queue[tuple[pa.Table, int] | _WorkerId],
        worker_idx: _WorkerId,
    ):
        try:
            while True:
                try:
                    x = result_handles.get_nowait()
                except queue.Empty:
                    break
                weight = x.estimated_uncompressed_size
                as_arrow = None
                if sem:
                    if weight > sem.initial_value:
                        # If the file is larger than the maximum size, we'll truncate it, so this file will be the only one being downloaded
                        weight = sem.initial_value
                    if weight > 0:
                        # No need to acquire the semaphore for empty tables
                        if not sem.acquire(weight):
                            raise RuntimeError("Failed to acquire semaphore for reading athena unload")
                        safe_set_gauge("chalk.athena.remaining_prefetch_bytes", sem.get_value())
                if as_arrow is None:
                    as_arrow = x.to_arrow(self._s3_filesystem())
                pa_table_queue.put((as_arrow, weight))
        finally:
            # At the end, putting the worker id to signal that this worker is done
            pa_table_queue.put(worker_idx)

    def _yield_from_result_handles(
        self,
        result_handles: queue.Queue[AthenaResultHandle],
        query_execution_parameters: QueryExecutionParameters,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
    ):
        yielded = False
        max_weight = query_execution_parameters.max_prefetch_size_bytes
        if max_weight <= 0:
            max_weight = None
        pa_table_queue: queue.Queue[tuple[pa.Table, int] | _WorkerId] = queue.Queue()
        sem = None if max_weight is None else MultiSemaphore(max_weight)
        assert query_execution_parameters.num_client_prefetch_threads >= 1

        futures = {
            _WorkerId(i): self.executor.submit(
                self._download_worker,
                result_handles,
                sem,
                pa_table_queue,
                _WorkerId(i),
            )
            for i in range(query_execution_parameters.num_client_prefetch_threads)
        }
        schema: pa.Schema | None = None

        while len(futures) > 0:
            x = pa_table_queue.get()
            if isinstance(x, int):
                futures.pop(x).result()
                continue
            tbl, weight = x
            if schema is None:
                schema = tbl.schema
            try:
                if len(tbl) == 0:
                    continue
                assert isinstance(tbl, pa.Table)
                features = columns_to_features(tbl.schema.names)
                yield self._postprocess_table(features, tbl)
                safe_incr("chalk.athena.downloaded_bytes", tbl.nbytes or 0)
                safe_incr("chalk.athena.downloaded_rows", tbl.num_rows or 0)
                yielded = True
            finally:
                # Releasing the semaphore post-yield to better respect the limit
                if sem is not None and weight > 0:
                    sem.release(weight)
                    safe_set_gauge("chalk.athena.remaining_prefetch_bytes", sem.get_value())
        if not yielded and query_execution_parameters.yield_empty_batches:
            if schema is not None:
                features = columns_to_features(schema.names)
                yield pa.RecordBatch.from_arrays(
                    arrays=[[] for _ in features],
                    names=[x.root_fqn for x in features.values()],
                )
                return

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_ATHENA_AWS_REGION_NAME, self.name, self.aws_region),
                create_integration_variable(_ATHENA_AWS_ACCESS_KEY_ID_NAME, self.name, self.aws_access_key_id),
                create_integration_variable(_ATHENA_AWS_ACCESS_KEY_SECRET_NAME, self.name, self.aws_access_key_secret),
                create_integration_variable(_ATHENA_S3_STAGING_DIR_NAME, self.name, self.s3_staging_dir),
                create_integration_variable(_ATHENA_ROLE_ARN_NAME, self.name, self.role_arn),
                create_integration_variable(_ATHENA_SCHEMA_NAME_NAME, self.name, self.schema_name),
                create_integration_variable(_ATHENA_CATALOG_NAME_NAME, self.name, self.catalog_name),
                create_integration_variable(_ATHENA_WORK_GROUP_NAME, self.name, self.work_group),
            ]
            if v is not None
        }



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Optional, Sequence, Type

import pyarrow as pa

from chalk.features._embedding.embedding_provider import EmbeddingProvider
from chalk.features._embedding.utils import ModelSpecs, create_fixedsize_with_nulls
from chalk.features._vector import Vector
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import openai
    import tiktoken
else:
    openai = None
    tiktoken = None

_MAX_INPUT_TOKENS = 8191
_MAX_BATCH_SIZE = 2048
_MODEL_SPECS = {
    "text-embedding-ada-002": ModelSpecs(
        default_dimensions=1536,
        validate_dimensions=lambda dim: dim == 1536,
        validation_message="Must be 1536.",
        requires_none_input=True,
    ),
    "text-embedding-3-small": ModelSpecs(
        default_dimensions=1536,
        validate_dimensions=lambda dim: 1 <= dim <= 1536,
        validation_message="Must be between 1 and 768 (inclusive).",
    ),
    "text-embedding-3-large": ModelSpecs(
        default_dimensions=3072,
        validate_dimensions=lambda dim: 1 <= dim <= 3072,
        validation_message="Must be between 1 and 3072 (inclusive).",
    ),
}


class OpenAIProvider(EmbeddingProvider):
    def __init__(self, model: str, dimensions: Optional[int] = None) -> None:
        super().__init__()
        try:
            global openai, tiktoken
            import openai
            import tiktoken
        except ImportError:
            raise missing_dependency_exception("chalkpy[openai]")
        if model not in _MODEL_SPECS:
            models_str = ", ".join(f"'{model}'" for model in _MODEL_SPECS)
            raise ValueError(f"Unsupported model '{model}' for OpenAI. The supported models are [{models_str}].")
        specs = _MODEL_SPECS[model]
        if dimensions is not None and not specs.validate_dimensions(dimensions):
            raise ValueError(f"Unsupported dimensions '{dimensions}' for model '{model}'. {specs.validation_message}")
        self.model = model
        self.input_dimensions = None if specs.requires_none_input else dimensions
        self.output_dimensions = dimensions if dimensions is not None else specs.default_dimensions

    @functools.cached_property
    def _async_client(self):
        assert openai is not None
        return openai.AsyncOpenAI()

    @functools.cached_property
    def _encoding(self):
        assert tiktoken is not None
        try:
            return tiktoken.encoding_for_model(self.model)
        except KeyError:
            # Use cl100k_base encoding by default
            return tiktoken.get_encoding("cl100k_base")

    def get_provider_name(self) -> str:
        return "openai"

    def get_model_name(self) -> str:
        return self.model

    def validate_input_schema(self, input_schema: Sequence[pa.DataType]) -> str | None:
        if len(input_schema) != 1:
            return f"OpenAI emeddings support only 1 input, but got {len(input_schema)} inputs"
        if input_schema[0] != pa.large_utf8():
            return f"OpenAI embeddings require a large_utf8() feature, but got a feature of type {input_schema[0]}"

    def _truncate_embedding_input(self, inp: str | None) -> str | None:
        if inp is None:
            return None
        if inp == "":
            return None
        input_tokens = self._encoding.encode(inp, allowed_special="all")
        if len(input_tokens) > _MAX_INPUT_TOKENS:
            return self._encoding.decode(input_tokens[:_MAX_INPUT_TOKENS])
        return inp

    def generate_embedding(self, input: pa.Table) -> pa.FixedSizeListArray:
        raise NotImplementedError("use async_generate_embedding instead")

    async def async_generate_embedding(self, input: pa.Table):
        inputs: list[str | None] = [self._truncate_embedding_input(i) for i in input.column(0).to_pylist()]
        # Step over `input` in chunks of MAX_BATCH_SIZE; the max openai array length
        for i in range(0, len(inputs), _MAX_BATCH_SIZE):
            chunked_input = inputs[i : i + _MAX_BATCH_SIZE]
            non_null_chunked_input: list[str] = []
            none_indices: set[int] = set()
            for idx, inp in enumerate(chunked_input):
                if inp is None:
                    none_indices.add(idx)
                else:
                    non_null_chunked_input.append(inp)
            try:
                if len(non_null_chunked_input) > 0:
                    if self.input_dimensions is not None:
                        response = await self._async_client.embeddings.create(
                            input=non_null_chunked_input, model=self.model, dimensions=self.input_dimensions
                        )
                    else:
                        response = await self._async_client.embeddings.create(
                            input=non_null_chunked_input, model=self.model
                        )
                    response_data = response.data
                else:
                    response_data = []
            except Exception as e:
                try:
                    distinct_types = {type(e) for e in chunked_input}
                except Exception:
                    distinct_types = None
                raise ValueError(
                    f"Failed to generate embeddings for inputs of length {len(chunked_input)}. "
                    + f"Found distinct types {distinct_types}. Error: {e}"
                ) from e
            values_with_nulls: list[Sequence[float] | None] = []
            response_position = 0
            for idx in range(len(chunked_input)):
                if idx in none_indices:
                    values_with_nulls.append(None)
                elif response_position < len(response_data):
                    values_with_nulls.append(response_data[response_position].embedding)
                    response_position += 1
                else:
                    raise ValueError(
                        f"Expected to find an embedding for input at position {idx}, but the response data was exhausted."
                    )
            yield create_fixedsize_with_nulls(values_with_nulls, self.output_dimensions)

    def get_vector_class(self) -> Type[Vector]:
        return Vector[self.output_dimensions]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import named_query_pb2 as chalk_dot_server_dot_v1_dot_named__query__pb2


class NamedQueryServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetAllNamedQueries = channel.unary_unary(
            "/chalk.server.v1.NamedQueryService/GetAllNamedQueries",
            request_serializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesResponse.FromString,
        )
        self.GetAllNamedQueriesActiveDeployment = channel.unary_unary(
            "/chalk.server.v1.NamedQueryService/GetAllNamedQueriesActiveDeployment",
            request_serializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesActiveDeploymentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesActiveDeploymentResponse.FromString,
        )
        self.GetNamedQueryByName = channel.unary_unary(
            "/chalk.server.v1.NamedQueryService/GetNamedQueryByName",
            request_serializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetNamedQueryByNameRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetNamedQueryByNameResponse.FromString,
        )


class NamedQueryServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetAllNamedQueries(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAllNamedQueriesActiveDeployment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetNamedQueryByName(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_NamedQueryServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetAllNamedQueries": grpc.unary_unary_rpc_method_handler(
            servicer.GetAllNamedQueries,
            request_deserializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesResponse.SerializeToString,
        ),
        "GetAllNamedQueriesActiveDeployment": grpc.unary_unary_rpc_method_handler(
            servicer.GetAllNamedQueriesActiveDeployment,
            request_deserializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesActiveDeploymentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesActiveDeploymentResponse.SerializeToString,
        ),
        "GetNamedQueryByName": grpc.unary_unary_rpc_method_handler(
            servicer.GetNamedQueryByName,
            request_deserializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetNamedQueryByNameRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_named__query__pb2.GetNamedQueryByNameResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.NamedQueryService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class NamedQueryService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetAllNamedQueries(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.NamedQueryService/GetAllNamedQueries",
            chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAllNamedQueriesActiveDeployment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.NamedQueryService/GetAllNamedQueriesActiveDeployment",
            chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesActiveDeploymentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_named__query__pb2.GetAllNamedQueriesActiveDeploymentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetNamedQueryByName(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.NamedQueryService/GetNamedQueryByName",
            chalk_dot_server_dot_v1_dot_named__query__pb2.GetNamedQueryByNameRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_named__query__pb2.GetNamedQueryByNameResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )
